# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Snapshots are wrappers around data frames. A snapshot ensures that row
identifier for the data frame are either -1 or a unique integer.
"""

from abc import ABCMeta
import numpy as np

from histore.document.reader import DocumentReader


class Document(metaclass=ABCMeta):
    """The document class wraps a pandas data frame. The data frame represents
    a snapshot in the history of the archived dataset. The document class
    provides identifier for each row in the data frame. These identifier are
    either derived using a primary key (i.e., from the cell values in each row
    of the data frame) or from the row index of the data frame. There exist
    different implementations of the abstract document class for each case.
    """
    def __init__(self, df, schema, rows):
        """Initialize the object properties. The abstract document class
        maintains a list of row (identifier, position) tuples in the order
        in which rows are being returned by the document reader (i.e., sorted
        by their row identifier).

        Raises a ValueError if the list of columns in the schema does not match
        the number of columns in the data frame.

        Parameters
        ----------
        df: pandas.DataFrame
            Pandas data frame representing the snapshot of an archived datset.
        schema: list(histore.document.schema.Column)
            List of columns in the document schema. Each column corresponds to
            a column in the data frame (based on list position). The schema
            columns provide the unique column identifier that are required by a
            document reader to generate document rows. An error is raised if
            the number of elements in the schema does not match the number of
            columns in the data frame.
        rows: list
            List of row (identifier, position) pairs. This list determines the
            order in which rows are returned by the document reader.

        Raises
        ------
        ValueError
        """
        # Raise an error if the number of elements in the schema does not match
        # the number of columns in the data frame.
        if len(df.columns) != len(schema):
            raise ValueError('invalid schema for data frame')
        self.df = df
        self.schema = schema
        self.rows = rows

    def reader(self):
        """Get reader for data frame rows ordered by their row identifier.

        Returns
        -------
        histore.document.reader.DocumentReader
        """
        return DocumentReader(df=self.df, schema=self.schema, rows=self.rows)


class PartialDocument(Document):
    """In a partial document the position information for rows in not aligned
    with the row positions in the full document. This class constructor aligns
    the row positions in the partial document with those for the full document
    given by a document index reader.

    The row positions in the original snapshot are used as the basis for
    alignment. That is, rows in the partial document will inherit the position
    of their counterpart in the snapshot document. New (unmatched) rows are
    appended at the end of the document.
    """
    def __init__(self, doc, row_index):
        """Align row positions in the partial document with row positions in
        the given row index reader.

        Parameters
        ----------
        doc: histore.document.base.Document
            Partial input document.
        row_index: histore.archive.reader.RowIndexReader
            Reader for row (identifier, position) tuples from the original
            snapshot version.
        """
        matched_rows, unmatched_rows = list(), list()
        rows = doc.rows
        readidx = 0
        orig_row = row_index.next()
        max_pos = -1
        while orig_row and readidx < len(rows):
            new_row = rows[readidx]
            if orig_row[0] < new_row[0]:
                # The original row is not included in the partial document.
                # Get information for next row in the original snapshot.
                orig_row = row_index.next()
            elif orig_row[0] > new_row[0]:
                # The row in the partial document is a new row that was not
                # present in the original document. Add the row identifier to
                # the list of unmatched rows.
                unmatched_rows.append(new_row[0])
                readidx += 1
            else:
                # Matched row. Adjust position of the row to the position of
                # the row in the original document.
                matched_rows.append(orig_row)
                orig_row = row_index.next()
                readidx += 1
            # Keep track of the maximum position in the original document.
            if orig_row and orig_row[1] > max_pos:
                max_pos = orig_row[1]
        # Finish reading the original document to get the maximum position
        # value.
        while orig_row:
            if orig_row[1] > max_pos:
                max_pos = orig_row[1]
            orig_row = row_index.next()
        # Append unmatched rows to the matched rows with adjusted row position.
        for key in unmatched_rows:
            max_pos += 1
            matched_rows.append((key, max_pos))
        # Append remaining rows in the partial document to the matched rows
        # list with adjusted position.
        while readidx < len(rows):
            max_pos += 1
            matched_rows.append((rows[readidx][0], max_pos))
            readidx += 1
        # Initialize the superclass with adjusted row positions.
        matched_rows.sort(key=lambda r: r[0])
        super(PartialDocument, self).__init__(
            df=doc.df,
            schema=doc.schema,
            rows=matched_rows
        )


# -- Documents for different row identifier types -----------------------------

class PKDocument(Document):
    """Snapshot document that derives row identifier from cell values in
    document rows using a primary key definition.
    """
    def __init__(self, df, schema, primary_key):
        """Initialize the object properties. Creates identifier for all rows
        in the document based on the given primary key.

        Parameters
        ----------
        df: pandas.DataFrame
            Pandas data frame representing the snapshot of an archived datset.
        schema: list(histore.document.schema.Column)
            List of columns in the document schema. Each column corresponds to
            a column in the data frame (based on list position). The schema
            columns provide the unique column identifier that are required by a
            document reader to generate document rows. An error is raised if
            the number of elements in the schema does not match the number of
            columns in the data frame.
        primary_key: string or list
            Name(s) of columns in the data frame schema from which the row keys
            are generated. If the primary key is not given the row index in the
            data frame is used as the row identifier. In the latter case all
            row identifier are expected to be integer or converted to integer.
            Negative values are assumed to represent new rows that will receive
            a new row identifier during merge.
        """
        # Generate row identifier from the column(s) in the primary key.
        if type(primary_key) not in [list, tuple]:
            # For convenience we ensure that we have a list of columns.
            primary_key = [primary_key]
        # For performance reasons we use two different implementations here
        # that distinguish between primary keys with a single column or with
        # multiple columns.
        rows = list()
        rowpos = 0
        if len(primary_key) == 1:
            keycol = primary_key[0]
            for _, values in df.iterrows():
                rows.append((values[keycol], rowpos))
                rowpos += 1
        else:
            for _, values in df.iterrows():
                rowid = tuple([values[c] for c in primary_key])
                rows.append((rowid, rowpos))
                rowpos += 1
        # Initialize the superclass.
        rows.sort(key=lambda r: r[0])
        super(PKDocument, self).__init__(df=df, schema=schema, rows=rows)


class RIDocument(Document):
    """Snapshot document that derives row identifier from the row index of the
    data frame. The identifier for new rows will be None.
    """
    def __init__(self, df, schema):
        """Initialize the object properties. Creates identifier for all rows
        in the document based on the data frame row index. Rows without an
        identifier are added at the end.

        Parameters
        ----------
        df: pandas.DataFrame
            Pandas data frame representing the snapshot of an archived datset.
        schema: list(histore.document.schema.Column)
            List of columns in the document schema. Each column corresponds to
            a column in the data frame (based on list position). The schema
            columns provide the unique column identifier that are required by a
            document reader to generate document rows. An error is raised if
            the number of elements in the schema does not match the number of
            columns in the data frame.
        """
        # Ensure that all row identifier are integers. Generate identifier for
        # those rows that have a non-integer index, None, or a negative value
        # as their index. These rows are considered new rows.
        rows, new_rows = list(), list()
        for pos in range(len(df.index)):
            rowid = df.index[pos]
            if not type(rowid) in [int, np.int64] or rowid < 0:
                new_rows.append((None, pos))
            else:
                rows.append((rowid, pos))
        # Initialize the superclass.
        rows = sorted(rows, key=lambda r: r[0]) + new_rows
        super(RIDocument, self).__init__(df=df, schema=schema, rows=rows)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Iterator for data frame rows. Provides a reader that allows to iterate over
the rows in a data frame sorted by their row identifier.
"""

import math

from histore.document.row import DocumentRow


class DocumentReader(object):
    """Reader for rows in a data frame. Reads rows in order defined by the
    sorted row identifer.
    """
    def __init__(self, df, schema, rows):
        """Initialize the wrapped data frame, the mapping of data frame columns
        to archive schema columns, and the information about row identifier and
        their position in the data frame.

        Parameters
        ----------
        df: pandas.DataFrame
            Pandas data frame representing the dataset snapshot.
        schema: list(histore.document.schema.Column)
            List of columns in the document schema. Each column corresponds to
            a column in the data frame (based on list position). The schema
            columns provide the unique column identifier that are required to
            generate document rows.
        rows: list
            List of (rowid, rowpos). The list is sorted by row identifier.
        """
        self.df = df
        self.schema = schema
        self.rows = rows
        # Maintain the position of the current row for the reader
        self.read_index = 0

    def has_next(self):
        """Test if the reader has more rows to read. If True the next() method
        will return the next row. Otherwise, the next() method will return
        None.

        Returns
        -------
        bool
        """
        return self.read_index < len(self.rows)

    def next(self):
        """Read the next row in the data frame. Returns None if the end of the
        data frame rows has been reached.

        Returns
        -------
        histore.document.row.DocumentRow
        """
        # Return None if the reader has no more rows.
        if not self.has_next():
            return None
        # Get identifier and position in the data frame for the next row that
        # is being returned.
        rowid, rowpos = self.rows[self.read_index]
        self.read_index += 1
        # Create mapping for values in the data frame row to the columns in the
        # archive schema.
        values = dict()
        row = self.df.iloc[rowpos]
        for i in range(len(self.schema)):
            val = row[i]
            try:
                val = None if math.isnan(val) else val
            except TypeError:
                pass
            values[self.schema[i].colid] = val
        return DocumentRow(key=rowid, pos=rowpos, values=values)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Provenance information about differences between two snapshots in a dataset
archive.
"""

from histore.archive.provenance.base import Provenance
from histore.archive.provenance.describe import (
    RowProvDescribe, SchemaProvDescribe
)


PROV_ROWS = 'ROWS'
PROV_SCHEMA = 'SCHEMA'


class SnapshotDiff(object):
    """The snapsshot difference object maintains provenance information about
    the difference betweeen two dataset snapshots. The changes are divided into
    changes to the dataset schema and the dataset rows.
    """
    def __init__(self, schema=None, rows=None):
        """Initialize provenance information about chages the dataset schema
        and row changes.

        Parameters
        ----------
        schema: histore.archive.provenance.base.Provenance
        rows: histore.archive.provenance.base.Provenance
        """
        if schema is None:
            schema = Provenance(descriptor=SchemaProvDescribe())
        if rows is None:
            rows = Provenance(descriptor=RowProvDescribe())
        self.prov = {
            PROV_SCHEMA: schema,
            PROV_ROWS: rows
        }

    def describe(self):
        """Print summary of the maintained provenance information."""
        print('Schema Changes\n==============')
        self.schema().describe()
        print('\nData Changes\n============')
        self.rows().describe()

    def rows(self):
        """Set of row objects that have changed between the two snapshots.

        Returns
        -------
        histore.archive.provenance.base.Provenance
        """
        return self.prov[PROV_ROWS]

    def schema(self):
        """Set of schema objects that have changed between the two snapshots.

        Returns
        -------
        histore.archive.provenance.base.Provenance
        """
        return self.prov[PROV_SCHEMA]



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Timestamps are sequences of snapshot (version) identifier that define when
a node was part of the archived document.
"""


class TimeInterval(object):
    """Interval of integer numbers. Defines that start and end of the interval.
    Ensures that start is lower or equal that end.
    """
    def __init__(self, start, end=None):
        """Initialize the interval from a given integer pair. If end is None it
        is assumed to be equal to start.

        Ensures that start is lower or equal than end (if given). Interval
        boundaries cannot be negative values.

        Parameters
        ----------
        start: int
            First version in the interval.
        end: int, optional
            Last version (inclusive) in the interval.
        """
        self.start = start
        self.end = end if end is not None else start
        if self.start < 0 or self.end < self.start:
            raise ValueError('invalid interval {}'.format(self))

    def __repr__(self):
        """Unambiguous string representation of this time interval.

        Returns
        -------
        string
        """
        return '[%i,%i]' % (self.start, self.end)

    def __str__(self):
        """Readable string representation of this time interval.

        Returns
        -------
        string
        """
        if self.start < self.end:
            return '%i-%i' % (self.start, self.end)
        else:
            return str(self.start)

    def contains(self, interval=None, version=None):
        """Returns true if this interval contains the given interval or value.

        Raises ValueError if both arguments are None or not None.

        Parameters
        ----------
        interval: histore.archive.timestamp.TimeInterval, optional
            Interval of version numbers.
        version: int, optional
            Single version number

        Returns
        -------
        bool

        Raises
        ------
        ValueError
        """
        if interval is not None and version is None:
            return self.start <= interval.start and self.end >= interval.end
        elif interval is None and version is not None:
            return self.start <= version <= self.end
        else:
            raise ValueError('invalid arguments')

    def is_equal(self, interval):
        """Returns True if the two intervals have equal start and end values.

        Parameters
        ----------
        interval: histore.archive.timestamp.TimeInterval
            Interval of version numbers.

        Returns
        -------
        bool
        """
        return self.start == interval.start and self.end == interval.end

    def overlap(self, interval):
        """Test if this interval overlaps with the given interval. To intervals
        overlap if they have at least one version in common.

        Parameters
        ----------
        interval: histore.archive.timestamp.TimeInterval
            Interval of version numbers.

        Returns
        -------
        bool
        """
        if self.start <= interval.start:
            return self.end >= interval.start
        elif self.start >= interval.start:
            return interval.end >= self.start
        return False


class Timestamp(object):
    """A timestamp is a sequence of integer numbers represented as a list of
    integer intervals.
    """
    def __init__(self, intervals=None, version=None):
        """Initialize the timestamp using an list of time intervals. Ensures
        that intervals are not adjacent or overlapping. Raises a ValueError if
        both arguments are not None.

        Parameters
        ----------
        intervals: list(histore.archive.timestamp.TimeInterval), optional
            List of intervals.
        version: int
            Unique version identifier.

        Raises
        ------
        ValueError
        """
        if intervals is not None and version is not None:
            raise ValueError('invalid arguments')
        if version is not None:
            self.intervals = [TimeInterval(version)]
        else:
            if intervals is None:
                self.intervals = list()
            elif isinstance(intervals, TimeInterval):
                self.intervals = [intervals]
            else:
                self.intervals = intervals
            # Validate the given list of intervals
            for i in range(1, len(self.intervals)):
                if self.intervals[i - 1].end + 1 >= self.intervals[i].start:
                    raise ValueError('ajacent or overlapping intervals')

    def __str__(self):
        """Get a string representation of the timestamp.

        Returns
        -------
        string
        """
        return ','.join([str(interval) for interval in self.intervals])

    def append(self, version):
        """Append the given version to the timestamp. Returns a new modified
        timestamp. Raises ValueError if the given value is lower or equal to
        the last version in the sequence.

        Parameters
        ----------
        version: int
            Version identifier.

        Returns
        -------
        histore.archive.timestamp.Timestamp
        """
        if len(self.intervals) > 0:
            last_interval = self.intervals[-1]
            if last_interval.end >= version:
                msg = 'cannot append {} to [{}]'
                raise ValueError(msg.format(version, self))
            elif version == last_interval.end + 1:
                interval = TimeInterval(start=last_interval.start, end=version)
                return Timestamp(intervals=self.intervals[:-1] + [interval])
            else:
                interval = TimeInterval(version)
                return Timestamp(intervals=self.intervals + [interval])
        else:
            return Timestamp(intervals=[TimeInterval(version)])

    def contains(self, version):
        """Returns True if the timestamp contains the given value.

        Parameters
        ----------
        version: int
            Version identifier.

        Returns
        -------
        bool
        """
        # Search from end of interval list since the majority of requests will
        # access more recent values (i.e., database versions).
        for interval in self.intervals[::-1]:
            if interval.contains(version=version):
                return True
            elif interval.end < version:
                return False
        return False

    @staticmethod
    def from_string(text):
        """Create a timestamp instance from a string representation as
        generated by the __str__ method.

        Parameters
        ----------
        text: string

        Returns
        -------
        histore.archive.timestamp.Timestamp
        """
        intervals = list()
        for token in text.strip().split(','):
            pos = token.find('-')
            if pos > 0:
                start = int(token[:pos])
                end = int(token[pos+1:])
                intervals.append(TimeInterval(start, end))
            else:
                intervals.append(TimeInterval(int(token)))
        return Timestamp(intervals=intervals)

    def is_empty(self):
        """Returns True if the timestamp is empty.

        Returns
        -------
        bool
        """
        return len(self.intervals) == 0

    def is_equal(self, timestamp):
        """Returns True if the two timestamps are equal, i.e., represent the
        same sequence of snapshots. Given that intervals cannot be consecutive,
        equality means (i) same number of intervals, and (ii) corresponding
        intervals have same start/end values.

        Parameters
        ----------
        timestamp: histore.archive.timestamp.Timestamp

        Returns
        -------
        bool
        """
        if len(self.intervals) == len(timestamp.intervals):
            for i in range(len(self.intervals)):
                if not self.intervals[i].is_equal(timestamp.intervals[i]):
                    return False
            return True
        return False

    def last_version(self):
        """Get the last version in the timestamp. If the timestamp is empty the
        result is None.

        Returns
        -------
        int
        """
        return self.intervals[-1].end if self.intervals else None



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Provenance information representing updates to scalar values. This class is
used to represent changes in dataset cell values, row positions, column names,
and column positions.
"""


class UpdateValue(object):
    """Provenance object representing an updated scalar value. Maintains a pair
    of old value (before update) and new value (after update).
    """
    def __init__(self, old_value, new_value):
        """Initialize the object properties.

        Parameters
        ----------
        old_value: scalar
            Cell value before update.
        new_value: scalar
            Cell value after update.
        """
        self.old_value = old_value
        self.new_value = new_value

    def __repr__(self):
        """Unambiguous string representation of the value update.

        Returns
        -------
        string
        """
        return '{} TO {}'.format(self.old_value, self.new_value)

    def values(self):
        """Get tuple containing old and new value.

        Returns
        -------
        scalar, scalar
        """
        return self.old_value, self.new_value



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Implementation of a basic archive store that maintains all data in main
memory. Archive information is not persisted.
"""

from histore.archive.schema import ArchiveSchema
from histore.archive.snapshot import SnapshotListing
from histore.archive.store.base import ArchiveStore
from histore.archive.store.mem.reader import BufferedReader
from histore.archive.store.mem.writer import ArchiveBuffer


class VolatileArchiveStore(ArchiveStore):
    """Archive store that keeps all archive related information in main memory.
    The store is volatile as no information is persisted on disk.
    """
    def __init__(self):
        """Initialize the archive archive components."""
        self.rows = list()
        self.schema = ArchiveSchema()
        self.snapshots = SnapshotListing()
        self.row_counter = 0

    def commit(self, schema, writer, snapshots):
        """Commit a new version of the dataset archive. The modified components
        of the archive are given as the three arguments of this method.

        Parameters
        ----------
        schema: histore.archive.schema.ArchiveSchema
            Schema history for the new archive version.
        writer: histore.archive.writer.ArchiveWriter
            Instance of the archive writer class returned by this store that
            was used to output the rows of the new archive version.
        snapshots: histore.archive.snapshot.SnapshotListing
            Modified list of snapshots in the new archive. The new archive
            version is the last entry in the list.
        """
        self.rows = writer.rows
        self.schema = schema
        self.snapshots = snapshots
        self.row_counter = writer.row_counter

    def is_empty(self):
        """True if the archive does not contain any snapshots yet.

        Returns
        -------
        bool
        """
        return self.snapshots.is_empty()

    def get_reader(self):
        """Get the row reader for this archive.

        Returns
        -------
        histore.archive.store.mem.BufferedReader
        """
        return BufferedReader(rows=self.rows)

    def get_schema(self):
        """Get the schema history for the archived dataset.

        Returns
        -------
        histore.archive.schema.ArchiveSchema
        """
        return self.schema

    def get_snapshots(self):
        """Get listing of all snapshots in the archive.

        Returns
        -------
        histore.archive.snapshot.SnapshotListing
        """
        return self.snapshots

    def get_writer(self):
        """Get a a new archive buffer to maintain rows for a new archive
        version.

        Returns
        -------
        histore.archive.store.mem.ArchiveBuffer
        """
        return ArchiveBuffer(row_counter=self.row_counter)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Collections of provenance information for archive snapshots. The provenance
represents the difference between two dataset snapshots with respect to
inserts, deletions, and updates.
"""

from histore.archive.provenance.describe import DefaultDescribe


"""Identifier for the three different provenance operator types."""
PROV_DELETE = 'DELETE'
PROV_INSERT = 'INSERT'
PROV_UPDATE = 'UPDATE'


class ProvOp(object):
    """Base class for provenance operators that represent changes between two
    versions of the same object. There are three different types of operators:
    (i) INSERT, (ii) DELETE, and (iii) UPDATE.
    """
    def __init__(self, type, key):
        """Initialize the operator type and the object key. Raises a ValueError
        if an invalid type is given.

        Parameters
        ----------
        type: string, enum=[PROV_DELETE, PROV_INSERT, PROV_UPDATE]
            Provenance operator type identifier.
        key: scalar or tuple
            Unique object identifier. This is either a column identifier, a row
            identifier, or a row key.

        Raises
        ------
        ValueError
        """
        if type not in [PROV_DELETE, PROV_INSERT, PROV_UPDATE]:
            raise ValueError('invalid provenance operator type'.format(type))
        self._type = type
        self.key = key

    def is_delete(self):
        """True if the operator represents object deletion.

        Returns
        -------
        bool
        """
        return self._type == PROV_DELETE

    def is_insert(self):
        """True if the operator represents object insertion.

        Returns
        -------
        bool
        """
        return self._type == PROV_INSERT

    def is_update(self):
        """True if the operator represents object update.

        Returns
        -------
        bool
        """
        return self._type == PROV_UPDATE


class Provenance(object):
    """Provenance information for a set of identifiable objects. Provenance for
    each object is represented by one of three operators: insert, delete, or
    update. The provenance represents the difference for the object between two
    dataset snapshots.
    """
    def __init__(self, insert=None, delete=None, update=None, descriptor=None):
        """Initialize the lists that maintain the different types of provenance
        operators. The optional provenance descriptor is used to print a
        summary of the maintained information.

        Parameters
        ----------
        insert: list, default=None
            List containing provenance insert operators for objects that have
            been inserted.
        delete: list, default=None
            List containing provenance delete operators for objects that have
            been deleted.
        update: list, default=None
            List containing provenance update operators for objects that have
            been udated.
        descriptor: histore.archive.provenance.describe.ProvDescriptor,
                    default=None
            Descriptor object that is used to print a summary of the maintained
            provenance information.
        """
        self.prov = {
            PROV_DELETE: delete if delete is not None else list(),
            PROV_INSERT: insert if insert is not None else list(),
            PROV_UPDATE: update if update is not None else list()
        }
        self.descriptor = descriptor if descriptor else DefaultDescribe()

    def add(self, prov):
        """Add provenance element to the collection.

        Parameters
        ----------
        prov: histore.archive.base.ProvOp
            Provenance operator.
        """
        if prov.is_delete():
            self.delete().append(prov)
        elif prov.is_insert():
            self.insert().append(prov)
        else:
            self.update().append(prov)

    def delete(self):
        """List containing provenance delete operators for objects that have
        been deleted.

        Returns
        -------
        list
        """
        return self.prov[PROV_DELETE]

    def describe(self):
        """Print summary of the maintained provenance information."""
        self.descriptor.describe(self)

    def insert(self):
        """List containing provenance insert operators for objects that have
        been inserted.

        Returns
        -------
        list
        """
        return self.prov[PROV_INSERT]

    def update(self):
        """List containing provenance update operators for objects that have
        been udated.

        Returns
        -------
        list
        """
        return self.prov[PROV_UPDATE]



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Abstract class that defines the interface for the archive writer."""

from abc import ABCMeta, abstractmethod

from histore.archive.row import ArchiveRow
from histore.archive.timestamp import Timestamp
from histore.archive.value import SingleVersionValue


class ArchiveWriter(metaclass=ABCMeta):
    """The abstract archive writer class defines the methods for adding rows
    to an output archive. The writer maintains an internal counter to assign
    unique identifier to document rows.
    """
    def __init__(self, row_counter):
        """Initialize the counter that is used to generate unique row
        identifier.

        Parameters
        ----------
        row_counter: int
            Counter that is used to generate unique internal row identifier.
            The current value of the counter is the value for the next unique
            identifier.
        """
        self.row_counter = row_counter

    @abstractmethod
    def write_archive_row(self, row):
        """Add the given row to a new archive version.

        Parameters
        ----------
        row: histore.archive.row.ArchiveRow
            Row in a new version of a dataset archive.
        """
        raise NotImplementedError()

    def write_document_row(self, row, version):
        """Add a given document row to a new archive version with the given
        identifier. Uses the internal row counter to assign a new unique row
        identifier.

        Parameters
        ----------
        row: histore.document.row.DocumentRow
            Row from an inout document (snapshot) that is being added to the
            archive snapshot for the given version.
        version: int
            Unique identifier for the snapshot version that the document row is
            added to.
        """
        # Create a new archive row with unique row identifier for the given
        # document row.
        ts = Timestamp(version=version)
        cells = dict()
        for colid, value in row.values.items():
            cells[colid] = SingleVersionValue(value=value, timestamp=ts)
        arch_row = ArchiveRow(
            rowid=self.row_counter,
            key=row.key if row.key is not None else self.row_counter,
            pos=SingleVersionValue(value=row.pos, timestamp=ts),
            cells=cells,
            timestamp=ts
        )
        # Increment the row counter and add the new row to the archive.
        self.row_counter += 1
        self.write_archive_row(arch_row)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Rows in an archived dataset. Each row has a timestamp, the index positions
of the row in the history of the dataset, and the cell values in the row
history.
"""

from histore.archive.provenance.row import DeleteRow, InsertRow, UpdateRow
from histore.archive.value import SingleVersionValue
from histore.archive.timestamp import Timestamp


class ArchiveRow(object):
    """Each row in an archived dataset has a unique row identifier. The row
    maintains the history of index positions that it had, and the history of
    all cell values.
    """
    def __init__(self, rowid, pos, cells, timestamp, key=None):
        """Initialize the row identifier and the objects that maintain history
        information about the row index positions and cell values.

        Parameters
        ----------
        rowid: int
            Unique internal row identifier
        pos: histore.archive.value.ArchiveValue
            Index positions for the row in the history of the dataset.
        cells: dict(int, histore.archive.value.ArchiveValue)
            Mapping of column identifier to cell values over the history of the
            dataset.
        timestamp: histore.archive.timestamp.Timestamp
            Sequence of version from the dataset history in which the row was
            present.
        key: int, string, or tuple, default=None
            Derived row key for matching and merging purposes. If the key is
            None the rowid is used as key
        """
        self.rowid = rowid
        self.cells = cells
        self.pos = pos
        self.timestamp = timestamp
        self.key = key if key is not None else rowid

    def __repr__(self):
        """Unambiguous string representation of the archive row.

        Returns
        -------
        string
        """
        return (
            '<ArchiveRow ('
            '\n\tid={}'
            '\n\tkey={}'
            '\n\ttimestamp=[{}]'
            '\n\tpos={}'
            '\n\tvalues={})>'
        ).format(
            self.rowid,
            self.key,
            str(self.timestamp),
            str(self.pos),
            self.cells
        )

    def at_version(self, version, columns, raise_error=True):
        """Get cell values and the index position for the row in the given
        version of the dataset. Returns the index position of the row and a
        list of cell values for all columns in the given list. The order of
        returned values matches the order of columns in the argument.

        Parameters
        ----------
        version: int
            Unique version identifier.
        columns: list(int)
            List of columns for which values are returned at the given version.
        raise_error: bool, default=True
            Flag that determines the behavior for cases where there is no value
            for the given version. If the flag is True an error is raised. If
            the flag is False None is returned.

        Returns
        -------
        int, dict
        """
        values = list()
        for colid in columns:
            # There may not be a value for the given column identifier. If the
            # raise error flag is True we will raise an error. Otherwise, None
            # is returned as the result.
            value = self.cells.get(colid)
            if value is None:
                if raise_error:
                    raise ValueError('unknown version %d' % version)
            else:
                value = value.at_version(version, raise_error=raise_error)
            values.append(value)
        return self.pos.at_version(version, raise_error=raise_error), values

    def comp(self, key):
        """Compare the identifier of the archive row to the given row key. The
        result is:

        - -1 if the archive key is lower that the given key or if the given
             key is None,
        -  0 if both keys are equal, or
        -  1 if the archive key is greater than the given key.

        Parameters
        ----------
        key: int, string, or tuple
            Key value of a document row.

        Returns
        -------
        int
        """
        if key is None or self.key < key:
            return -1
        elif self.key > key:
            return 1
        else:
            return 0

    def diff(self, original_version, new_version):
        """Get provenance information representing the difference for this
        row between the original version and a new version.

        The three main types of changes that the row can experience are:

        - InsertRow: The row did not exists in the original version but in the
          new version.
        - DeleteRow: The row did exist in the original version but not in the
          new version.
        - UpdateRow: Row updates can be comprised of two types of changes:
                     (i) The row position changed
                     (ii) One or more row values were updated

        If the row has no changes between both versions the result is None.

        Parameters
        ----------
        original_version: int
            Unique identifier for the original version.
        new_version: int
            Unique identifier for the version that the original version is
            compared against.

        Returns
        -------
        histore.archive.provenance.row.RowOp
        """
        exists_in_orig = self.timestamp.contains(original_version)
        exists_in_new = self.timestamp.contains(new_version)
        if exists_in_orig and not exists_in_new:
            # Row has been deleted
            return DeleteRow(self.key)
        elif not exists_in_orig and exists_in_new:
            # Row has been inserted
            return InsertRow(self.key)
        elif exists_in_orig and exists_in_new:
            # Check if row has been updated.
            upd_cells = dict()
            for colid, value in self.cells.items():
                upd_val = value.diff(original_version, new_version)
                if upd_val is not None:
                    upd_cells[colid] = upd_val
            upd_pos = self.pos.diff(original_version, new_version)
            if upd_pos is not None or len(upd_cells) > 0:
                return UpdateRow(
                    key=self.key,
                    cells=upd_cells,
                    position=upd_pos
                )
        # Row has not changed
        return None

    def extend(self, version, origin):
        """Extend the timestamp of the row and all cell values and the index
        position that were valid at the given version of origin.

        Parameters
        ----------
        version: int
            Unique identifier of the new version.
        origin: int
            Identifier of the version of origin.

        Returns
        -------
        histore.archive.row.ArchiveRow
        """
        # If the row was not present at the given version of origin the object
        # is returned unchanged.
        if not self.timestamp.contains(origin):
            return self
        # Extend all cell values that were valid at the origin.
        ext_cells = dict()
        for colid, value in self.cells.items():
            ext_cells[colid] = value.extend(version=version, origin=origin)
        # Return an extended copy of the row. the row position is assumed to be
        # unchanged.
        return ArchiveRow(
            rowid=self.rowid,
            key=self.key,
            pos=self.pos.extend(version=version, origin=origin),
            cells=ext_cells,
            timestamp=self.timestamp.append(version)
        )

    def merge(self, values, pos, version, unchanged_cells=None, origin=None):
        """Create a new version of the dataset row for a given set of cell
        values and a specified index position. The set of extend cells
        identifies those columns that are not present in the value dictionary
        but that remain valid in the new version with respect to the given
        source version (origin).

        Values is a dictionary that maps column identifier to cell values in
        the new row. Columns that are not present in the dictionary the have
        either be deleted or remain unchanged. In the latter case, these
        columns are included in the set of unchanged columns. Columns that
        are unchanged are keep the value that they had in the given source
        version (origin).

        Parameters
        ----------
        values: dict
            Cell values in the new row. The dictionary is a mapping from column
            identifier to cell values.
        pos: int
            Index position of the row in the dataset snapshot.
        version: int
            Identifier of the new row version.
        unchanged_cells: set, default=None
            Set of identifier for columns whose cell values are not included in
            the values dictionary but that remain unchanged with respect to the
            specified source version (origin).
        origin: int, default=None
            Version that the row values originate from. Cell that remain
            unchanged have the timestamp extended for the version the was
            present the version of origin.

        Returns
        -------
        histore.archive.row.ArchiveRow
        """
        # Create a modified dictionary of cell values for all columns.
        history = dict()
        for colid, value in values.items():
            if colid in self.cells:
                cell = self.cells[colid]
                cell = cell.merge(value=value, version=version)
            else:
                cell = SingleVersionValue(
                    value=value,
                    timestamp=Timestamp(version=version)
                )
            history[colid] = cell
        # For missing columns either add them as they are (e.g., if the column
        # has been deleted) or if they are included in the unchanged set append
        # the new version to the timestamp for the value that was valid at the
        # source version.
        unchanged = unchanged_cells if unchanged_cells is not None else set()
        for colid, cell in self.cells.items():
            if colid not in history:
                if colid in unchanged:
                    cell = cell.extend(
                        version=version,
                        origin=origin
                    )
                history[colid] = cell
        # Return a modified archive row.
        return ArchiveRow(
            rowid=self.rowid,
            key=self.key,
            pos=self.pos.merge(value=pos, version=version),
            cells=history,
            timestamp=self.timestamp.append(version)
        )



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Columns in histore data frames have a unique identifier and a column name.
The column class extends the Python String class to be able to be used as a
column value in a Pandas data frame.
"""


class Column(str):
    """Columns in openclean data frames are subclasses of Python strings that
    contain a unique column identifier. This implementation is based on:
    https://bytes.com/topic/python/answers/32098-my-experiences-subclassing-string

    The order of creation is that the __new__ method is called which returns
    the object then __init__ is called.
    """
    def __new__(cls, colid, name, *args, **keywargs):
        """Initialize the String object with the given column name. Ignore the
        column identifier.

        Parameters
        ----------
        colid: int
            Unique column identifier
        name: string
            Column name
        """
        return str.__new__(cls, str(name))

    def __init__(self, colid, name):
        """Initialize the unique column identifier. The column name has already
        been initialized by the __new__ method that is called prior to the
        __init__ method.

        Parameters
        ----------
        colid: int
            Unique column identifier
        name: string
            Column name
        """
        self.colid = colid



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Columns in an archive schema all have unique identifier together with
timestamped names and positions.
"""

from histore.archive.provenance.column import (
    DeleteColumn, InsertColumn, UpdateColumn
)
from histore.archive.timestamp import Timestamp
from histore.archive.value import SingleVersionValue
from histore.document.schema import Column


"""Constant identifier for the three different schema matching modes."""
MATCH_ID = 'idonly'
MATCH_IDNAME = 'idname'
MATCH_NAME = 'nameonly'


class ArchiveColumn(object):
    """Archive columns have a unique identifier together with a timestamped
    name and schema position.
    """
    def __init__(self, identifier, name, pos, timestamp):
        """initialize the object properties.

        Parameters
        ----------
        identifier: int
            Unique column identifier
        name: histore.archive.value.ArchiveValue
            History of names for the column.
        pos: histore.archive.value.ArchiveValue
            History of schema positions for the column.
        timestamp: histore.archive.timestamp.Timestamp
            Timestamp for the full history of the column.
        """
        self.identifier = identifier
        self.name = name
        self.pos = pos
        self.timestamp = timestamp

    def __repr__(self):
        """Unambiguous string representation of the archive column.

        Returns
        -------
        string
        """
        template = (
            '<ArchiveColumn (\n'
            '\tid=%d\n'
            '\tname=%s\n'
            '\tpos=%s\n'
            '\tts=%s\n)>'
        )
        return template % (
            self.identifier,
            str(self.name),
            str(self.pos),
            str(self.timestamp)
        )

    def at_version(self, version):
        """Get the name and schema position for the column in the given
        version of a dataset. Returns a tuple of (name, pos).

        Parameters
        ----------
        version: int
            Unique version identifier.

        Returns
        -------
        (string, int)
        """
        return self.name.at_version(version), self.pos.at_version(version)

    def diff(self, original_version, new_version):
        """Get provenance information representing the difference for this
        column between the original version and a new version.

        The three main types of changes that the column can experience are:

        - InsertColumn: The column did not exists in the original version but
          in the new version.
        - DeleteColumn: The column did exist in the original version but not in
          the new version.
        - UpdateColumn: Column updates can be comprised of two types of
          changes:
            (i) The column position changed
            (ii) The column name changed

        If the column has no changes between both versions the result is None.

        Parameters
        ----------
        original_version: int
            Unique identifier for the original version.
        new_version: int
            Unique identifier for the version that the original version is
            compared against.

        Returns
        -------
        histore.archive.provenance.column.ColumnOp
        """
        exists_in_orig = self.timestamp.contains(original_version)
        exists_in_new = self.timestamp.contains(new_version)
        if exists_in_orig and not exists_in_new:
            # Column has been deleted
            return DeleteColumn(self.identifier)
        elif not exists_in_orig and exists_in_new:
            # Column has been inserted
            return InsertColumn(self.identifier)
        elif exists_in_orig and exists_in_new:
            # Check if column name or position has been updated.
            upd_name = self.name.diff(original_version, new_version)
            upd_pos = self.pos.diff(original_version, new_version)
            if upd_pos is not None or upd_name is not None:
                return UpdateColumn(
                    key=self.identifier,
                    name=upd_name,
                    position=upd_pos
                )
        # Column has not changed
        return None

    def merge(self, name, pos, version):
        """Create a modified version of the column for the given version. Adds
        the name and index position to the respective history of the column.
        Returns a modified copy of the column.

        Parameters
        ----------
        name: string
            Column name.
        pos: int
            Position of the column in the snapshot schema.

        Returns
        -------
        histore.archive.schema.ArchiveColumn
        """
        return ArchiveColumn(
            identifier=self.identifier,
            name=self.name.merge(value=name, version=version),
            pos=self.pos.merge(value=pos, version=version),
            timestamp=self.timestamp.append(version)
        )


class ArchiveSchema(object):
    """The archive schema maintains an index of columns in the history of a
    dataset. Each column has a unique identifier together with timestamped
    names and positions.
    """
    def __init__(self, columns=None):
        """Initialize the index of columns in the schema. Raises a ValueError
        if the identifier of columns are not unique.

        Parameters
        ----------
        columns: dict or list, default=None
            Dictionary or list of archive columns. In a dictionary, columns are
            indexed by their unique identifier. If a list is given it will be
            converted to a dictionary.
        """
        if columns is None:
            # Assert that columns is not None
            self.columns = dict()
        elif isinstance(columns, list):
            # If columns is a list convert it to a dictionary
            self.columns = dict()
            for c in columns:
                if c.identifier in self.columns:
                    raise ValueError('duplicate identifier %d' % c.identifier)
                self.columns[c.identifier] = c
        elif isinstance(columns, dict):
            self.columns = columns
        else:
            raise ValueError('expected list or dict of columns')

    def __iter__(self):
        """Make column list iterable."""
        return iter(self.columns.values())

    def __repr__(self):
        """Unambiguous string representation of the archive schema.

        Returns
        -------
        string
        """
        return '\n'.join(str(c) for c in self.columns.values())

    def at_version(self, version):
        """Get the schema for the snapshot at the given version of a dataset.
        Returns a list of snapshot columns in their original order.

        Parameters
        ----------
        version: int
            Unique version identifier.

        Returns
        -------
        list(histore.document.schema.Column)
        """
        # Get a list of (identifier, name, pos) for all columns in the snapshot
        # schema at the given version.
        cols = list()
        for col in self.columns.values():
            if col.timestamp.contains(version):
                name, pos = col.at_version(version)
                cols.append((col.identifier, name, pos))
        # Sort columns based on their position and return a list of snapshot
        # columns.
        cols.sort(key=lambda x: x[2])
        return [Column(colid=id, name=name) for id, name, _ in cols]

    def merge(
        self, columns, version, matching=MATCH_IDNAME, renamed=None,
        renamed_to=True, partial=False, origin=None
    ):
        """Add the given snapshot columns to the schema history for a dataset.
        Expects a list of strings or identifiable column objects. If column
        objects are given their identifier must match an existing identifier
        in the schema and all identifier have to be unique. Column names that
        are given as string-only are considered to be to be new columns.

        If the column matching mode involves matching by name or if the partial
        flag is True the version of origin has to be given as well. When
        matching by name the columns in the given list are matched against
        columns in the schema of the origin version. Columns that are unmatched
        are considered new columns. If either the list of columns of the origin
        schema contains duplicate column names matching by name is not possible
        and an error is raised.

        If the partial flag is true columns from the origin schema that are
        not matched by columns in the given list (either by name or by their
        identifier) are added to the resulting schema. An attempt is made to
        keep the order of columns as close as possible to the order of columns
        in the origin schema.

        Returns a modified archive schema, a list of column objects
        corresponding to the columns in the given list, and a list of column
        objects from the original snapshot schema that were unmatched.

        Parameters
        ----------
        columns: list
            List of column names in their original order. Column names can
            either be strings of snapshot columns.
        version: int
            Identifier of the new version.
        matching: string, default='idname'
            Match mode for columns. Excepts one of three modes:
            - idonly: The columns in the schema of the comitted data frame are
            matched against columns in the archive schema by their identifier.
            Assumes that columns in the data frame schema are instances of the
            class histore.document.schema.Column.
            - nameonly: Columns in the commited data frame schema are matched
            by name against the columns in the schema of the snapshot that is
            identified by origin.
            - idname: Match columns of type histore.document.schema.Column
            first against the columns in the archive schema. Match remaining
            columns by name against the schema of the snapshot that is
            identified by origin.
        renamed: dict, default=None
            Optional mapping of columns that have been renamed. Maps the new
            column name to the original name.
        renamed_to: bool, default=True
            Flag that determines the semantics of the mapping in the renamed
            dictionary. By default a mapping from the original column name
            (i.e., the dictionary key) to the new column name (the dictionary
            value) is assumed. If the flag is False a mapping from the new
            column name to the original column name is assumed.
        partial: bool, default=False
            If True the list of columns is assumed partial. All columns from
            the snapshot schema that is specified by origin that are not
            matched by any column in the input list are added to the schema
            for the new snapshot before merging the snapshot schema into the
            archive.
        origin: int, default=None
            Version identifier of the original column against which the given
            column list is matched.

        Returns
        -------
        histore.archive.schema.ArchiveSchema, list, list

        Raises
        ------
        ValueError
        """
        # Initialize the original schema that is needed when matching by name
        # or if a partial schema is given. Create mappings for column names to
        # archive columns and column position as well as for column identifier
        # to column and position.
        if origin is not None:
            orig_schema = self.at_version(origin)
            archive_index, orig_index = column_index(
                archive_schema=self.columns,
                orig_schema=orig_schema
            )
        else:
            orig_schema, archive_index = list(), dict()
        # Get alignment of data frame columns with archive columns by column
        # identifier. If matching is by name only we pass an empty dictionary
        # as archive schema so that all columns are unmatched. the result is a
        # list of tuples containing the data frame column and the matched
        # archive column (or None if unmatched).
        archive_cols = self.columns if matching != MATCH_NAME else dict()
        matches = columnid_match(df_cols=columns, archive_cols=archive_cols)
        if matching != MATCH_ID:
            # Find matches for columns by name in the snapshot schema. Then
            # modify the columns list by replacing entries with their
            # matched schema columns (if matched).
            matches = match_columns(
                columns=matches,
                schema_index=archive_index,
                renamed=renamed,
                renamed_to=renamed_to
            )
        # Create the new archive schema. Merge matched columns. Add columns
        # in the data frame schema that are unmatched. Keep track of unmatched
        # columns in the original schema.
        arch_columns = dict()
        matched_columns = list()
        colids = max(self.columns) + 1 if len(self.columns) > 0 else 0
        nextpos = len(orig_schema) if partial else 0
        for i in range(len(matches)):
            dfcol, archcol = matches[i]
            name = str(dfcol)
            if archcol is not None:
                # The snapshot column was matched with an existing archive
                # column. The name of these columns, however, may be different
                # if the column was renamed. Merge the snapshot column into the
                # matched archive column. Get the column position either from
                # the origin index or the nextpos counter.
                if partial:
                    if archcol.identifier in orig_index:
                        # Get column position. Then remove the matched column
                        # from the index.
                        _, pos = orig_index[archcol.identifier]
                        del orig_index[archcol.identifier]
                    else:
                        pos = nextpos
                        nextpos += 1
                else:
                    pos = i
                # Create a new column object with the
                # name from the new snapshot and the identifier from the
                # original schema.
                archcol = archcol.merge(name=name, pos=pos, version=version)
                col = Column(colid=archcol.identifier, name=name)
            else:
                # The snapshot column was not matched. Create a new column and
                # add it to the archive. Append the column to a partial schema
                # or use the original index as the column position.
                if partial:
                    pos = nextpos
                    nextpos += 1
                else:
                    pos = i
                ts = Timestamp(version=version)
                archcol = ArchiveColumn(
                    identifier=colids,
                    name=SingleVersionValue(value=name, timestamp=ts),
                    pos=SingleVersionValue(value=pos, timestamp=ts),
                    timestamp=ts
                )
                nextpos += 1
                col = Column(colid=colids, name=name)
                colids += 1
            if archcol.identifier in arch_columns:
                raise ValueError('duplicate column %d' % archcol.identifier)
            arch_columns[archcol.identifier] = archcol
            matched_columns.append(col)
        # Add original columns that are not in the modified schema
        unmatched_columns = list()
        for colid, archcol in self.columns.items():
            if colid not in arch_columns:
                if partial and colid in orig_index:
                    # Missing column from a partial schema.
                    dfcol, pos = orig_index[colid]
                    col = str(dfcol)
                    archcol = archcol.merge(name=col, pos=pos, version=version)
                    unmatched_columns.append(dfcol)
                arch_columns[colid] = archcol
        # Return a modified archive schema, the resulting list of matched
        # columns and the list of columns that remain unchanged.
        schema = ArchiveSchema(columns=arch_columns)
        return schema, matched_columns, unmatched_columns


# -- Helper Methods -----------------------------------------------------------

def columnid_match(df_cols, archive_cols):
    """Match columns in the given data frame schema against columns in the
    archive schema by their identifier. Returns a list of tuples where each
    tuple contains the data frame column and the matched archive schema column.
    If a data frame column is unmatched the second tuple value is None. The
    order of elements in the result corresponds to the order of columns in the
    input data frame schema.

    Parameters
    ----------
    df_cols: list
        List of strings or column objects from a pandas data frame schema.
    archive_cols: dict(histore.archive.schema.ArchiveColumn)
        Dictionary of columns in the schema of a dataset archive.

    Returns
    -------
    string or histore.document.schema.Column,
    histore.archive.schema.ArchiveColumn or None
    """
    alignment = list()
    for pos in range(len(df_cols)):
        dfcol = df_cols[pos]
        if isinstance(dfcol, Column):
            # Get archive column with the identifier if it exists.
            alignment.append((dfcol, archive_cols.get(dfcol.colid)))
        else:
            # Strings cannot be matched by column indentifier.
            alignment.append((dfcol, None))
    return alignment


def match_columns(columns, schema_index, renamed=None, renamed_to=True):
    """Compute matching of given list of 'id-aligned' columns to columns in the
    given snapshot schema. Columns are matched by name. If either list contains
    duplicate column names an error is raised.

    Returns a list of tuples (column, match) where column is an element from
    the list of columns and match is either an element from the schema or None
    for unmatched columns. The order of elements in the result corresponds to
    the order of elements in the column list.

    Parameters
    ----------
    columns: list((string, histore.archive.schema.ArchiveColumn or None)
        List of column names.
    schema_index: dict(string: histore.archive.schema.ArchiveColumn)
        Dictionary mapping of column names (in a snapshot schema) to their
        respective archive column.
    renamed: dict, default=None
        Optional mapping of columns that have been renamed. Maps the new column
        name to the original name.
    renamed_to: bool, default=True
        Flag that determines the semantics of the mapping in the renamed
        dictionary. By default a mapping from the original column name (i.e.,
        the dictionary key) to the new column name (the dictionary value) is
        assumed. If the flag is False a mapping from the new column name to the
        original column name is assumed.

    Returns
    -------
    list(string, histore.document.schema.Column)

    Raises
    ------
    ValueError
    """
    # Ensure that the semantics of the renamed mapping is from the new name to
    # their original name. This makes it easier to lookup the column names in
    # the new snapshot schema to find their original column name.
    if renamed is None:
        renamed_from = dict()
    elif renamed_to:
        # Flip mapping of renamed columns.
        renamed_from = dict()
        for key, val in renamed.items():
            if val in renamed_from:
                raise ValueError('multiple columns renamed to same name')
            renamed_from[val] = key
    else:
        renamed_from = renamed
    # Match column names against the schema index. Keep track if column names
    # identifiy duplciates. Ensure that the result is a valid mapping where no
    # two names from the column list have been mapped to the same schema
    # element.
    matches = list()
    matched_schema_columns = set()
    names = set()
    for df_col, arch_col in columns:
        if arch_col is not None:
            matches.append((df_col, arch_col))
            continue
        name = str(df_col)
        if name in names:
            raise ValueError('duplicate column name %s' % (str(name)))
        names.add(name)
        # Find matching name in the schema index. Consider the fact that the
        # column may have been renamed from an original column.
        match = schema_index.get(renamed_from.get(name, name))
        if match is not None:
            if match.identifier in matched_schema_columns:
                raise ValueError('multiple matches for %d' % match.colid)
            matched_schema_columns.add(match.identifier)
        matches.append((name, match))
    return matches


def column_index(archive_schema, orig_schema):
    """Create a mapping of names for the snapshot schema to their respective
    column in the archive schema. In addition, a mapping of identifier for
    columns in the snapshot schema to the column object at its schema position
    is created. Returns a tuple with the name mapping and the column identifier
    mapping.

    Parameters
    ----------
    archive_schema: dict(int: histore.archive.schema.ArchiveColumn)
        Dictionary mapping of column identifier to archive columns.
    orig_schema: list(histore.document.schema.Column)
        List of columns in a snapshot schema

    Returns
    -------
    dict, dict
    """
    archive_index = dict()
    orig_index = dict()
    for pos in range(len(orig_schema)):
        col = orig_schema[pos]
        if col in archive_index:
            raise ValueError('cannot match by name with duplicates')
        archive_index[col] = archive_schema[col.colid]
        orig_index[col.colid] = (col, pos)
    return archive_index, orig_index



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Reader for archive rows that are maintained as a Json array in a file on the
file system.
"""

import json
import os

from histore.archive.reader import ArchiveReader
from histore.archive.serialize.default import DefaultSerializer

import histore.util as util


class ArchiveFileReader(ArchiveReader):
    """Reader for rows in a dataset archive. Reads rows in ascending order of
    their identifier.
    """
    def __init__(self, filename, serializer=None, compression=None):
        """
        Parameters
        ----------
        filename: string
            Path to the output file.
        serializer: histore.archive.serialize.ArchiveSerializer, default=None
            Implementation of the archive serializer interface that is used to
            serialize rows that are written to file.
        compression: string, default=None
            String representing the compression mode for the output file.
        """
        self.serializer = serializer if serializer else DefaultSerializer()
        # The archive is empty if the given file does not exists. In this case
        # the buffer will remain empty,
        self.buffer = None
        if os.path.isfile(filename):
            self.fin = util.inputstream(filename, compression=compression)
            # Read the first two lines in the output file. The first line is
            # expected to be '['. The second line is either ']' (for an empty
            # archive) or contain the first row.
            if self.fin.readline() != '[':
                raise ValueError('invalid input file {}'.format(filename))
            self.next()
        else:
            self.fin = None

    def has_next(self):
        """The next row for the reader is the current buffer value. If the
        buffer is empty there is no next row.

        Returns
        -------
        bool
        """
        return self.buffer is not None

    def next(self):
        """Read the next row in the dataset archive. Returns None if the end of
        the archive rows has been reached.

        Returns
        -------
        histore.archive.row.ArchiveRow
        """
        # The end of the rows has been reached if the input stream is None.
        if not self.fin:
            return None
        # The return value is the current buffer value.
        result = self.buffer
        # Read the next line. If the line equals the closing array bracket the
        # end of the archive has been reached.
        line = self.fin.readline()
        if line == ']':
            # The end of the archive was reached. Close the input file.
            self.buffer = None
            self.fin.close()
            self.fin = None
        else:
            # Remove trailing comma for list array elements before the final
            # element.
            if line.endswith(','):
                line = line[:-1]
            self.buffer = self.serializer.deserialize_row(json.loads(line))
        # Return the previous buffer value as the result.
        return result



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Store for archives that are materialized as files on the file system. Each
archive is maintained in a separate folder on disk. For each archive two
different files are created.

- metadata.dat: Json file containing the archive schema, snapshot information,
  and the row counter.
- rows.dat: Data fie containing the archive rows.
"""

import json
import os
import shutil

from histore.archive.schema import ArchiveSchema
from histore.archive.serialize.default import DefaultSerializer
from histore.archive.snapshot import SnapshotListing
from histore.archive.store.base import ArchiveStore
from histore.archive.store.fs.reader import ArchiveFileReader
from histore.archive.store.fs.writer import ArchiveFileWriter

import histore.util as util


"""Element labels for metadata serialization."""
META_SCHEMA = 'schema'
META_SNAPSHOTS = 'snapshots'
META_ROWCOUNT = 'rowcount'


class ArchiveFileStore(ArchiveStore):
    """Archive store that maintains archive data in files on the file system.
    Maintains a copy of the schema and the snapshot descriptors in memory for
    faster access.
    """
    def __init__(
        self, basedir, replace=False, serializer=None, compression=None
    ):
        """Initialize the archive archive components.

        Parameters
        ----------
        basedir: string
            Path to the base directory for archive files. If the directory does
            not exist it will be created.
        replace: boolean, default=False
            Do not read existing files in the base directory to initialize the
            archive. Treat the archive as being empty instead if True.
        serializer: histore.archive.serialize.ArchiveSerializer, default=None
            Implementation of the archive serializer interface that is used to
            serialize rows that are written to file.
        compression: string, default=None
            String representing the compression mode. Only te data file will be
            compressed. the metadata file is always storesd as plain text.
        """
        self.basedir = util.createdir(basedir)
        self.serializer = serializer if serializer else DefaultSerializer()
        self.compression = compression
        # Initialize the file names
        self.datafile = os.path.join(self.basedir, 'rows.dat')
        self.metafile = os.path.join(self.basedir, 'metadata.dat')
        self.tmpdatafile = os.path.join(self.basedir, 'rows.tmp')
        self.tmpmetafile = os.path.join(self.basedir, 'metadata.tmp')
        if not replace and os.path.isfile(self.metafile):
            # Read schema and snapshot information from disk if the metadata
            # file exists.
            with open(self.metafile, 'r') as f:
                doc = json.load(f)
            # Deserialize schema columns.
            columns = list()
            for c in doc[META_SCHEMA]:
                columns.append(self.serializer.deserialize_column(c))
            self.schema = ArchiveSchema(columns=columns)
            # Deserialize snapshot descriptors
            snapshots = list()
            for s in doc[META_SNAPSHOTS]:
                snapshots.append(self.serializer.deserialize_snapshot(s))
            self.snapshots = SnapshotListing(snapshots=snapshots)
            # Deserialize row counter.
            self.row_counter = doc[META_ROWCOUNT]
        else:
            # Create an empty archive.
            self.schema = ArchiveSchema()
            self.snapshots = SnapshotListing()
            self.row_counter = 0
            # Remove any previous files that may exist in the base folder.
            for f in [self.datafile, self.metafile]:
                if os.path.isfile(f):
                    os.remove(f)

    def commit(self, schema, writer, snapshots):
        """Commit a new version of the dataset archive. The modified components
        of the archive are given as the three arguments of this method.

        Parameters
        ----------
        schema: histore.archive.schema.ArchiveSchema
            Schema history for the new archive version.
        writer: histore.archive.writer.ArchiveWriter
            Instance of the archive writer class returned by this store that
            was used to output the rows of the new archive version.
        snapshots: histore.archive.snapshot.SnapshotListing
            Modified list of snapshots in the new archive. The new archive
            version is the last entry in the list.
        """
        # Write temporary metadata file to disk (to ensure that write does not
        # raise an error).
        encoder = self.serializer
        doc = {
            META_SCHEMA: [encoder.serialize_column(c) for c in schema],
            META_SNAPSHOTS: [encoder.serialize_snapshot(s) for s in snapshots],
            META_ROWCOUNT: writer.row_counter
        }
        with open(self.tmpmetafile, 'w') as f:
            json.dump(doc, f)
        # Close the archive writer.
        writer.close()
        # Replace existing files with temporary files for new archive version.
        # This is the point of no return.
        # TODO: Instead of moving files we could delete (or keep) previous
        # files as backup.
        shutil.move(src=self.tmpmetafile, dst=self.metafile)
        shutil.move(src=self.tmpdatafile, dst=self.datafile)
        # Udate the cached objects
        self.schema = schema
        self.snapshots = snapshots
        self.row_counter = writer.row_counter

    def is_empty(self):
        """True if the archive does not contain any snapshots yet.

        Returns
        -------
        bool
        """
        return self.snapshots.is_empty()

    def get_reader(self):
        """Get the row reader for this archive.

        Returns
        -------
        histore.archive.store.mem.BufferedReader
        """
        return ArchiveFileReader(
            filename=self.datafile,
            serializer=self.serializer,
            compression=self.compression
        )

    def get_schema(self):
        """Get the schema history for the archived dataset.

        Returns
        -------
        histore.archive.schema.ArchiveSchema
        """
        return self.schema

    def get_snapshots(self):
        """Get listing of all snapshots in the archive.

        Returns
        -------
        histore.archive.snapshot.SnapshotListing
        """
        return self.snapshots

    def get_writer(self):
        """Get a a new archive buffer to maintain rows for a new archive
        version.

        Returns
        -------
        histore.archive.store.mem.ArchiveBuffer
        """
        # Use a temporary file in the base directory to write the new archive
        # before commit.
        return ArchiveFileWriter(
            filename=self.tmpdatafile,
            row_counter=self.row_counter,
            serializer=self.serializer,
            compression=self.compression
        )



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Values in a database archive. Represent cells in dataset rows as well as
information about the dataset schema (i.e., column name and position). Archived
values may either have multiple timestamped values, e.g., the different values
in a row cell over the history of the data set, or there is only one single
value, e.g., if the cell never changed over the history of the dataset.
"""

from abc import ABCMeta, abstractmethod

from histore.archive.provenance.value import UpdateValue
from histore.archive.timestamp import Timestamp


class ArchiveValue(metaclass=ABCMeta):
    """The archive value represents the history of a cell in a tabular dataset.
    """
    @abstractmethod
    def at_version(self, version, raise_error=True):
        """Get cell value for the given version. Raises ValueError if the cell
        does not have a value for the given version and the raise error flag is
        set tot True. If the flag is false None is returned instead.

        Parameters
        ----------
        version: int
            Unique version identifier.
        raise_error: bool, default=True
            Flag that determines the behavior for cases where there is no value
            for the given version. If the flag is True an error is raised. If
            the flag is False None is returned.

        Returns
        -------
        scalar

        Raises
        ------
        ValueError
        """
        raise NotImplementedError()

    def diff(self, original_version, new_version):
        """Get provenance information representing the difference for this
        value between the original version and a new version. If the value
        was the same for both versions the result is None.

        Parameters
        ----------
        original_version: int
            Unique identifier for the original version.
        new_version: int
            Unique identifier for the version that the original version is
            compared against.

        Returns
        -------
        histore.archive.provenance.value.UpdateValue
        """
        old_value = self.at_version(original_version, raise_error=False)
        new_value = self.at_version(new_version, raise_error=False)
        if old_value != new_value:
            return UpdateValue(old_value=old_value, new_value=new_value)
        # The value has not changed.
        return None

    @abstractmethod
    def extend(self, version, origin):
        """Extend the timestamp of the value that was valid at the given source
        version with the new version identifier. If no value was valid at the
        given version of origin the value is returned unchanged.

        Parameters
        ----------
        version: int
            Unique identifier of the new version.
        origin: int
            Unique identifier of the version for which the valid values
            timestamp is being extended.

        Returns
        -------
        histore.archive.value.ArchiveValue

        Raises
        ------
        ValueError
        """
        raise NotImplementedError()

    def is_multi_version(self):
        """Helper method to get the type of an archive value. Values can either
        be single version values or multi-version values.

        Returns
        -------
        bool
        """
        return not self.is_single_version()

    @abstractmethod
    def is_single_version(self):
        """Helper method to get the type of an archive value. Values can either
        be single version values or multi-version values.

        Returns
        -------
        bool
        """
        raise NotImplementedError()

    @abstractmethod
    def merge(self, value, version):
        """Add value for the given version into the cell history. Returns a
        modified copy of the value.

        Parameters
        ----------
        value: scalar
            Scalar cell value.
        version: int
            Unique version identifier.

        Returns
        -------
        hisotre.archive.value.ArchiveValue
        """
        raise NotImplementedError()


class SingleVersionValue(ArchiveValue):
    """Single archive value that has never changed over to history of the
    dataset that contains it.
    """
    def __init__(self, value, timestamp):
        """Initialize the archived value and the timestamp.

        Parameters
        ----------
        value: scalar
            Scalar value.
        timestamp: histore.archive.timestamp.Timestamp
            Timestamp of the archived value.
        """
        self.value = value
        self.timestamp = timestamp

    def __repr__(self):
        """Unambiguous string representation of the single version value.

        Returns
        -------
        string
        """
        return '({} [{}])'.format(self.value, str(self.timestamp))

    def at_version(self, version, raise_error=True):
        """Get value for the given version. If the given version is not
        included in the timestamp of the value an error is raised (if the raise
        error flag is True) or None is returned (otherwise).

        Parameters
        ----------
        version: int
            Unique version identifier.
        raise_error: bool, default=True
            Flag that determines the behavior for cases where there is no value
            for the given version. If the flag is True an error is raised. If
            the flag is False None is returned.

        Returns
        -------
        scalar

        Raises
        ------
        ValueError
        """
        if self.timestamp.contains(version):
            # Return value if it was present at the given version.
            return self.value
        elif raise_error:
            # Value was not present at the given version. Raise error if the
            # respective flag is True.
            raise ValueError('unknown version %d' % version)
        else:
            # Return None as default value for the given version.
            return None

    def extend(self, version, origin):
        """Extend the timestamp of the value with the given version identifier.

        Parameters
        ----------
        version: int
            Unique identifier of the new version.
        origin: int
            Unique identifier of the version for which the valid values
            timestamp is being extended.

        Returns
        -------
        histore.archive.value.ArchiveValue

        Raises
        ------
        ValueError
        """
        # Return the object unchanged if the version of origin is not contained
        # in the current timestamp.
        if not self.timestamp.contains(origin):
            return self
        # Return a modified value that extends the validity of the timestamp
        # for the given version.
        return SingleVersionValue(
            value=self.value,
            timestamp=self.timestamp.append(version)
        )

    def is_single_version(self):
        """Return True for the single version value class.

        Returns
        -------
        bool
        """
        return True

    def merge(self, value, version):
        """Add value for the given version into the cell history. Depending on
        whether the value is different from previous cell value this either
        returns the current object itself with an extended timestamp or a new
        multi-valued object.

        Parameters
        ----------
        value: scalar
            Scalar cell value.
        version: int
            Unique version identifier.

        Returns
        -------
        hisotre.archive.value.ArchiveValue
        """
        if self.value == value:
            # Extend the timestamp of the value.
            return SingleVersionValue(
                value=self.value,
                timestamp=self.timestamp.append(version)
            )
        # Return a multi-valued object.
        return MultiVersionValue(values=[
            TimestampedValue(value=self.value, timestamp=self.timestamp),
            TimestampedValue(value=value, timestamp=Timestamp(version=version))
        ])


class MultiVersionValue(ArchiveValue):
    """A multi-versiond archive value has multiple different values over the
    history of the dataset. Each value is associated with its own timestamp.
    """
    def __init__(self, values):
        """Initialize the list of timestamped values in the value history.

        Parameters
        ----------
        values: list(histore.archive.value.TimestampedValue)
            Values in the cell history.
        """
        self.values = values

    def __repr__(self):
        """Unambiguous string representation of the multi-versioned value.

        Returns
        -------
        string
        """
        return '({})'.format(', '.join(str(v) for v in self.values))

    def at_version(self, version, raise_error=True):
        """Get value for the given version. Raises ValueError if the value does
        not have an entry that was valid for the given version.

        Parameters
        ----------
        version: int
            Unique version identifier.
        raise_error: bool, default=True
            Flag that determines the behavior for cases where there is no value
            for the given version. If the flag is True an error is raised. If
            the flag is False None is returned.

        Returns
        -------
        scalar

        Raises
        ------
        ValueError
        """
        for val in self.values:
            if val.timestamp.contains(version):
                return val.value
        # No value for the version.
        if raise_error:
            raise ValueError('unknown version %d' % version)
        return None

    def extend(self, version, origin):
        """Extend the timestamp of the value that was valid at the given source
        version with the new version identifier. If no value was valid at the
        given version of origin the value is returned unchanged.

        Parameters
        ----------
        version: int
            Unique identifier of the new version.
        origin: int
            Unique identifier of the version for which the valid values
            timestamp is being extended.

        Returns
        -------
        histore.archive.value.ArchiveValue

        Raises
        ------
        ValueError
        """
        modified_values = list(self.values)
        for i in range(len(modified_values)):
            val = modified_values[i]
            if val.timestamp.contains(origin):
                # Found the value that was valid at the source version. Modify
                # the associated timestamp and return immediately.
                modified_values[i] = val.append(version)
                return MultiVersionValue(values=modified_values)
        # No value for the source version. Return the value unchanged.
        return self

    def is_single_version(self):
        """Return False for the single version value class.

        Returns
        -------
        bool
        """
        return False

    def merge(self, value, version):
        """Add value for the given version into the cell history. Returns a
        modified copy of the value.

        Parameters
        ----------
        value: scalar
            Scalar cell value.
        version: int
            Unique version identifier.

        Returns
        -------
        hisotre.archive.value.MultiVersionValue
        """
        # Create modified list of values in the cell history.
        cell_history = list()
        # Check if the value existed in the history of the cell. If a match is
        # found the timestamp of the value is adjusted accordingly.
        existed = False
        for val in self.values:
            if val.value == value:
                existed = True
                val = val.append(version)
            cell_history.append(val)
        # If the value did not exist in any prior version we need to add it to
        # the cell history.
        if not existed:
            ts = Timestamp(version=version)
            cell_history.append(TimestampedValue(value=value, timestamp=ts))
        return MultiVersionValue(values=cell_history)


class TimestampedValue:
    """The timestamped value is a wrapper around a scalar cell value and the
    associated timestamp.
    """
    def __init__(self, value, timestamp):
        """Initialize the object properties.

        Parameters
        ----------
        value: scalar
            Scalar cell value.
        timestamp: histore.archive.timestamp.Timestamp
            Sequence of version numbers when the value was present in a dataset
            row cell.
        """
        self.value = value
        self.timestamp = timestamp

    def __repr__(self):
        """Unambiguous string representation of the timestamped value.

        Returns
        -------
        string
        """
        return '{} [{}]'.format(self.value, str(self.timestamp))

    def append(self, version):
        """Return a modified copy if the timestamped value where the given
        version has been appended to to timestamp.

        Parameters
        ----------
        version: int
            Unique version identifier.

        Returns
        -------
        histore.archive.value.TimestampedValue
        """
        return TimestampedValue(
            value=self.value,
            timestamp=self.timestamp.append(version)
        )



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Classes to maintain information about dataset snapshots in an archive."""

import histore.util as util


class Snapshot(object):
    """Descriptor for archive snapshots. Contains all the meta-data about a
    single snapshot in a dataset archive.
    """
    def __init__(
        self, version, valid_time, transaction_time=None, description=None
    ):
        """Initialize the snapshot meta-data.

        Parameters
        ----------
        version: int
            Unique version identifier for the snapshot.
        valid_time: datetime.datetime
            Timestamp when the snapshot was first valid. A snapshot is valid
            until the valid time of the next snapshot in the archive.
        transaction_time: datetime.datetime, default=None
            Timestamp when the snapshot was inserted into the archive. If the
            timestamp is not given the current time is used.
        description: string, default=None
            Optional user-provided description for the snapshot.
        """
        self.version = version
        self.valid_time = valid_time
        if transaction_time is None:
            transaction_time = util.utc_now()
        self.transaction_time = transaction_time
        self.description = description if description is not None else ''

    def __repr__(self):
        """Unambiguous string representation of the archive snapshot
        descriptor.

        Returns
        -------
        string
        """
        return "<Snapshot (version={} description='{}' at={})>".format(
            self.version,
            self.description,
            str(util.to_localtime(self.transaction_time))
        )


class SnapshotListing(object):
    """Listing of snapshot descriptors for snapshots in an archive. Ensures
    that the order of snapshots reflects the order of their version numbers
    and valid times.
    """
    def __init__(self, snapshots=None):
        """Initialize the list of snapshot descriptors. Raises a ValueError if
        the order of snapshots in the list does not reflect the order of their
        version numbers and valid times.

        Parameters
        ----------
        snapshots: histore.archive.snapshot.Snapshot, default=None
            List of snapshots descriptors for snapshots in an archive.

        Raises
        ------
        ValueError
        """
        self.snapshots = snapshots if snapshots is not None else list()
        # Validate the order of version numbers and valid times.
        if len(self.snapshots) > 1:
            prev = self.snapshots[0]
            for i in range(1, len(self.snapshots)):
                s = self.snapshots[i]
                if s.version <= prev.version:
                    msg = 'invalid version sequence{} ->{}'
                    raise ValueError(msg.format(prev.version, s.version))
                if s.valid_time < prev.valid_time:
                    msg = 'invalid time sequence {} -> {}'
                    raise ValueError(msg.format(prev.valid_time, s.valid_time))
                prev = s

    def __iter__(self):
        """Make the snapshot list iterable.

        Returns
        -------
        iterator
        """
        return iter(self.snapshots)

    def __len__(self):
        """Get number of snapshots in the listing.

        Returns
        -------
        int
        """
        return len(self.snapshots)

    def append(self, valid_time=None, description=None):
        """Add a new version to the given listing. This will return a modified
        version listing with the new snapshot as the last element.

        The snapshot version will be one greater that the last version. If the
        snapshot list is empty the first version will be zero. If the valid
        time is not given the transaction time is used as valid time.

        Raises a valueError if the new snapshot violates the order of valid
        times in the list.

        Parameters
        ----------
        valid_time: datetime.datetime, default=None
            Timestamp when the snapshot was first valid. A snapshot is valid
            until the valid time of the next snapshot in the archive.
        description: string, default=None
            Optional user-provided description for the snapshot.

        Returns
        -------
        histore.archive.snapshot.SnapshotListing

        Raises
        ------
        ValueError
        """
        if not self.snapshots:
            version = 0
        else:
            version = self.snapshots[-1].version + 1
        transaction_time = util.utc_now()
        valid_time = valid_time if valid_time is not None else transaction_time
        s = Snapshot(
            version=version,
            valid_time=valid_time,
            transaction_time=transaction_time,
            description=description
        )
        return SnapshotListing(snapshots=self.snapshots + [s])

    def at_time(self, ts):
        """Get the snapshot that was valid at the given time. A snapshot is
        considered valid starting at its valit_time until the next timestamp.
        The last snapshot in the list is considered valid until infinity.

        Returns None if the given timestamp is before the valid time of the
        first snapshot.

        Parameters
        ----------
        ts: datetime.datetime
            Timestamp in UTC time.

        Returns
        -------
        histore.archive.snapshot.Snapshot
        """
        if not self.snapshots:
            # Return None if there are no snapshots yet.
            return None
        # Ensure that the first snapshot was valid at or before the given
        # timestamp.
        s1 = self.snapshots[0]
        if ts < s1.valid_time:
            # There is no snapshot at the given valid time.
            return None
        # Find the snapshot that was valid after the given timestamp and return
        # the previous snapshot.
        for i in range(1, len(self.snapshots)):
            s2 = self.snapshots[i]
            if ts < s2.valid_time:
                return s1
            s1 = s2
        # The last snapshot was valid at or before the given timestamp.
        return s1

    def has_version(self, version):
        """Check if the given version identifier references an existing
        snapshot in the listing.

        Parameters
        ----------
        version: int
            Unique version identifier.

        Returns
        -------
        bool
        """
        for s in self.snapshots:
            if s.version == version:
                return True
        return False

    def is_empty(self):
        """Shortcut to test if the list of snapshots is empty.

        Returns
        -------
        bool
        """
        return not self.snapshots

    def last_snapshot(self):
        """Shortcut to get the last snapshot in the list. The result is None if
        the listing is empty.

        Returns
        -------
        histore.archive.snapshot.Snapshot
        """
        if not self.snapshots:
            return None
        return self.snapshots[-1]



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Default object serializer for dataset archives."""

from histore.archive.row import ArchiveRow
from histore.archive.schema import ArchiveColumn
from histore.archive.snapshot import Snapshot
from histore.archive.timestamp import TimeInterval, Timestamp
from histore.archive.value import (
    MultiVersionValue, SingleVersionValue, TimestampedValue
)
from histore.archive.serialize.base import ArchiveSerializer

import histore.util as util


class DefaultSerializer(ArchiveSerializer):
    """Implementation of the archive object serializer. This is the default
    serializer used by HISTORE.
    """
    def __init__(
        self, timestamp='t', pos='p', name='n', cells='c', value='v',
        key='k', rowid='r', colid='c', version='v', valid_time='vt',
        transaction_time='tt', description='d'
    ):
        """Initialize the labels for elements used in the serialized objects.
        By default short labels are used to reduce storage overhead.

        Parameters
        ----------
        timestamp: string, default='t'
            Element label for timestamp objects.
        pos: string, default='p'
            Element label for objects index position values.
        name: string, default='n'
            Element label for column names.
        cells: string, default='c'
            Element label for row cell values.
        value: string, default='v'
            Element label for timestamped values.
        key: string, default='k'
            Element label for row key values.
        rowid: string, default='r'
            Element label for arcive row identifier.
        colid: string, default='c'
            Element label for column identifier.
        version: string, default='v'
            Element label for snapshot version number.
        valid_time: string, default='vt'
            Element label for snapshot valid time.
        transaction_time: string, default='tt'
            Element label for snapshot transaction time.
        description: string, default='d'
            Element label for snapshot descriptions.
        """
        self.timestamp = timestamp
        self.pos = pos
        self.name = name
        self.cells = cells
        self.value = value
        self.key = key
        self.rowid = rowid
        self.colid = colid
        self.version = version
        self.valid_time = valid_time
        self.transaction_time = transaction_time
        self.description = description

    def deserialize_column(self, obj):
        """Get archive schema column instance from a serialized object.

        Parameters
        ----------
        obj: dict
            Serialized archive row object.

        Returns
        -------
        histore.archive.schema.ArchiveColumn
        """
        ts = self.deserialize_timestamp(obj[self.timestamp])
        return ArchiveColumn(
            identifier=obj[self.colid],
            name=self.deserialize_value(obj=obj[self.name], ts=ts),
            pos=self.deserialize_value(obj=obj[self.pos], ts=ts),
            timestamp=ts
        )

    def serialize_column(self, column):
        """Get serialization for an archive schema column. Creates a dictionary
        with elements for the column identifier, name, the index position, and
        the column timestamp.

        Parameters
        ----------
        column: histore.archive.schema.ArchiveColumn
            Archive column object that is being serialized.

        Returns
        -------
        dict
        """
        ts = column.timestamp
        return {
            self.colid: column.identifier,
            self.name: self.serialize_value(value=column.name, ts=ts),
            self.pos: self.serialize_value(value=column.pos, ts=ts),
            self.timestamp: self.serialize_timestamp(ts)
        }

    def deserialize_row(self, obj):
        """Get archive row instance from a serialized object. Expects a
        dictionary as created by the serialize_row method.

        Parameters
        ----------
        obj: dict
            Serialized archive row object.

        Returns
        -------
        histore.archive.row.ArchiveRow
        """
        ts = self.deserialize_timestamp(obj[self.timestamp])
        rowid = obj[self.rowid]
        pos = self.deserialize_value(obj=obj[self.pos], ts=ts)
        cells = dict()
        for colid, value in obj[self.cells].items():
            cells[int(colid)] = self.deserialize_value(obj=value, ts=ts)
        if self.key in obj:
            key = obj[self.key]
            if isinstance(key, list):
                key = tuple(key)
        else:
            key = None
        return ArchiveRow(
            rowid=rowid,
            pos=pos,
            cells=cells,
            timestamp=ts,
            key=key
        )

    def serialize_row(self, row):
        """Get serialization for an archive row object.

        Parameters
        ----------
        row: histore.archive.row.ArchiveRow
            Archive row object that is being serialized.

        Returns
        -------
        dict
        """
        ts = row.timestamp
        cells = dict()
        for colid, value in row.cells.items():
            cells[colid] = self.serialize_value(value=value, ts=ts)
        obj = {
            self.rowid: row.rowid,
            self.timestamp: self.serialize_timestamp(ts),
            self.pos: self.serialize_value(value=row.pos, ts=ts),
            self.cells: cells
        }
        if row.key != row.rowid:
            obj[self.key] = row.key
        return obj

    def deserialize_snapshot(self, obj):
        """Get snapshot descriptor instance from a serialized object.

        Parameters
        ----------
        obj: dict
            Serialized snapshot descriptor object.

        Returns
        -------
        histore.archive.snapshot.Snapshot
        """
        if self.description in obj:
            description = obj[self.description]
        else:
            description = None
        return Snapshot(
            version=obj[self.version],
            valid_time=util.to_datetime(obj[self.valid_time]),
            transaction_time=util.to_datetime(obj[self.transaction_time]),
            description=description
        )

    def serialize_snapshot(self, snapshot):
        """Get serialization for an archive snapshot descriptor.

        Parameters
        ----------
        snapshot: histore.archive.snapshot.Snapshot
            Archive snapshot descriptor that is being serialized.

        Returns
        -------
        dict
        """
        obj = {
            self.version: snapshot.version,
            self.valid_time: snapshot.valid_time.isoformat(),
            self.transaction_time: snapshot.transaction_time.isoformat()
        }
        if snapshot.description is not None:
            obj[self.description] = snapshot.description
        return obj

    def deserialize_timestamp(self, obj):
        """Get timestamp instance from serialization.

        Parameters
        ----------
        obj: list
            Serialized timestamp that is a list of 2-dimensional lists
            representing the time intervals.

        Returns
        -------
        histore.archive.timestamp.Timestamp
        """
        intervals = [TimeInterval(start=i[0], end=i[1]) for i in obj]
        return Timestamp(intervals=intervals)

    def serialize_timestamp(self, ts):
        """Get serialization for atimestamp object. A timestamp is serialized
        as a list of 2-dimensional lists that contain the start and end version
        of the time interval.

        Parameters
        ----------
        ts: histore.archive.timestamp.Timestamp
            Archive timestamp object is a list of time intervals.

        Returns
        -------
        list
        """
        return [[interval.start, interval.end] for interval in ts.intervals]

    def deserialize_value(self, obj, ts):
        """Get timestamped value object from a serialization object. Expects a
        dictionary for single version values or a list of dictionaries for
        multi-version values.

        Parameters
        ----------
        obj: list or dict
            Serialized timestamped values as generatred by the serialize_value
            method.
        ts: histore.archive.timestamp.Timestamp
            Timestamp of the parent object.

        Returns
        -------
        string, list, or dict
        """
        if isinstance(obj, dict):
            if self.timestamp in obj:
                ts = self.deserialize_timestamp(obj[self.timestamp])
            return SingleVersionValue(value=obj[self.value], timestamp=ts)
        elif isinstance(obj, list):
            values = []
            for v in obj:
                values.append(
                    TimestampedValue(
                        value=v[self.value],
                        timestamp=self.deserialize_timestamp(v[self.timestamp])
                    )
                )
            return MultiVersionValue(values=values)
        raise ValueError('invalid serialization object')

    def serialize_value(self, value, ts):
        """Get serialization for a timestamp value. Depending on whether the
        value is a single version value or a multi-version value either a
        dictionary (single version) or a list of dictionaries (multi-version)
        is returned.

        Parameters
        ----------
        value: histore.archive.value.ArchiveValue
            Timestamped value in a dataset archive.
        ts: histore.archive.timestamp.Timestamp
            Timestamp of the parent object.

        Returns
        -------
        list or dict
        """
        if value.is_single_version():
            # A single version value is a dictionary containing the timestamped
            # value and an optional timestamp (if different from the parent
            # timestamp).
            obj = dict()
            if not value.timestamp.is_equal(ts):
                obj[self.timestamp] = self.serialize_timestamp(value.timestamp)
            obj[self.value] = value.value
        else:
            # A multi-version value is serialized as a list of single version
            # values each with their own timestamp.
            obj = [{
                self.timestamp: self.serialize_timestamp(v.timestamp),
                self.value: v.value
            } for v in value.values]
        return obj



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Abstract class for object serialization and deserialization."""

from abc import ABCMeta, abstractmethod


class ArchiveSerializer(metaclass=ABCMeta):
    """The serializer is used the serialize and deserialize components of
    dataset archives. All serializations are either lists or dictionaries.
    """
    @abstractmethod
    def deserialize_column(self, obj):
        """Get archive schema column instance from a serialized object. The
        type of the object depends on the serializer implementation. It is
        assumed that the object has been serialized using the respective method
        from the same implementing class.

        Parameters
        ----------
        obj: string, list, or dict
            Serialized archive row object.

        Returns
        -------
        histore.archive.schema.ArchiveColumn
        """
        raise NotImplementedError()

    @abstractmethod
    def serialize_column(self, column):
        """Get serialization for an archive schema column.

        Parameters
        ----------
        column: histore.archive.schema.ArchiveColumn
            Archive column object that is being serialized.

        Returns
        -------
        string, list, or dict
        """
        raise NotImplementedError()

    @abstractmethod
    def deserialize_row(self, obj):
        """Get archive row instance from a serialized object. The type of the
        object depends on the serializer implementation. It is assumed that
        the object has been serialized using the respective method from
        the same implementing class.

        Parameters
        ----------
        obj: string, list, or dict
            Serialized archive row object.

        Returns
        -------
        histore.archive.row.ArchiveRow
        """
        raise NotImplementedError()

    @abstractmethod
    def serialize_row(self, row):
        """Get serialization for an archive row object.

        Parameters
        ----------
        row: histore.archive.row.ArchiveRow
            Archive row object that is being serialized.

        Returns
        -------
        string, list, or dict
        """
        raise NotImplementedError()

    @abstractmethod
    def deserialize_snapshot(self, obj):
        """Get snapshot descriptor instance from a serialized object. The type
        of the object depends on the serializer implementation. It is assumed
        that the object has been serialized using the respective method from
        the same implementing class.

        Parameters
        ----------
        obj: string, list, or dict
            Serialized snapshot descriptor object.

        Returns
        -------
        histore.archive.snapshot.Snapshot
        """
        raise NotImplementedError()

    @abstractmethod
    def serialize_snapshot(self, snapshot):
        """Get serialization for an archive snapshot descriptor.

        Parameters
        ----------
        snapshot: histore.archive.snapshot.Snapshot
            Archive snapshot descriptor that is being serialized.

        Returns
        -------
        string, list, or dict
        """
        raise NotImplementedError()

    @abstractmethod
    def deserialize_timestamp(self, obj):
        """Get timestamp instance from serialized object. The type of the
        object depends on the serializer implementation. It is assumed that
        the object has been serialized using the respective method from
        the same implementing class.

        Parameters
        ----------
        obj: string, list, or dict
            Serialized timestamp object.

        Returns
        -------
        histore.archive.timestamp.Timestamp
        """
        raise NotImplementedError()

    @abstractmethod
    def serialize_timestamp(self, ts):
        """Get serialization for a timestamp object.

        Parameters
        ----------
        ts: histore.archive.timestamp.Timestamp
            Archive timestamp object is a list of time intervals.

        Returns
        -------
        string, list, or dict
        """
        raise NotImplementedError()

    @abstractmethod
    def deserialize_value(self, obj, ts):
        """Get timestamped value object from a serialization object. The type
        of the object depends on the serializer implementation. It is assumed
        that the object has been serialized using the respective method from
        the same implementing class.

        Parameters
        ----------
        ts: histore.archive.timestamp.Timestamp
            Archive timestamp object is a list of time intervals.

        Returns
        -------
        string, list, or dict
        """
        raise NotImplementedError()

    @abstractmethod
    def serialize_value(self, value, ts):
        """Get serialization for a timestamp value.

        Parameters
        ----------
        value: histore.archive.value.ArchiveValue
            Timestamped value in a dataset archive.
        ts: histore.archive.timestamp.Timestamp
            Timestamp of the parent object.

        Returns
        -------
        string, list, or dict
        """
        raise NotImplementedError()



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Implementation of the archive reader for the in-memory archive."""

from histore.archive.reader import ArchiveReader


class BufferedReader(ArchiveReader):
    """Reader for a list of archive rows that are kept in main memory."""
    def __init__(self, rows):
        """Initialize the list of rows that the reader will return.

        Parameters
        ----------
        rows: list(histore.archive.row.ArchiveRow)
            List of rows in the input stream.
        """
        self.rows = rows
        # Maintain a read index that points to the next row in the input
        # stream.
        self.read_index = 0

    def has_next(self):
        """Test if the reader has more rows to read. If True the next() method
        will return the next row. Otherwise, the next() method will return
        None.

        Returns
        -------
        bool
        """
        return self.read_index < len(self.rows)

    def next(self):
        """Read the next row in the dataset archive. Returns None if the end of
        the archive rows has been reached.

        Returns
        -------
        histore.archive.row.ArchiveRow
        """
        # Return None if the reader has no more rows.
        if not self.has_next():
            return None
        # Get the next row that the read index points to. Advance the read
        # index before returning that row..
        row = self.rows[self.read_index]
        self.read_index += 1
        return row



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Class representing provenance information for dataset columns in an archive.
Column provenance information is represented as operators that describe the
differences in a single dataset column between two dataset snapshots.
"""

import histore.archive.provenance.base as prov


class DeleteColumn(prov.ProvOp):
    """Provenance class representing the column deletion operator."""
    def __init__(self, key):
        """Initialize the operator type in the super class and the column
        identifier.

        Parameters
        ----------
        key: int
            Unique column identifier.
        """
        super(DeleteColumn, self).__init__(type=prov.PROV_DELETE, key=key)


class InsertColumn(prov.ProvOp):
    """Provenance class representing the column insert operator."""
    def __init__(self, key):
        """Initialize the operator type in the super class and the column
        identifier.

        Parameters
        ----------
        key: int
            Unique column identifier.
        """
        super(InsertColumn, self).__init__(type=prov.PROV_INSERT, key=key)


class UpdateColumn(prov.ProvOp):
    """Provenance class representing the column update operator. The column
    update operator maintains additional information about potential changes in
    the column position and the column name.
    """
    def __init__(self, key, name=None, position=None):
        """Initialize information about the parts of the column that have
        changed. This includes changes to the column name and the column
        position. Both types of changes are optional but at least one of them
        has to be given.

        Parameters
        ----------
        key: int
            Unique column identifier.
        name: histore.archive.provenance.value.UpdateValue, default=None
            Updated value representing the old and new column name.
        position: histore.archive.provenance.value.UpdateValue, default=None
            Updated value representing the old and new position of the column
            in the dataset schema.

        Raises
        ------
        ValueError
        """
        # Raise an error if both arguments are None
        if not name and not position:
            raise ValueError('no column changes given')
        super(UpdateColumn, self).__init__(type=prov.PROV_UPDATE, key=key)
        self.name = name
        self.position = position

    def updated_name(self):
        """Get updated column name.

        Returns
        -------
        dict
        """
        return self.name

    def updated_position(self):
        """Get updated column position.

        Returns
        -------
        histore.archive.provenance.value.UpdateValue
        """
        return self.position



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Implementation of the archive writer for the in-memory archive."""


from histore.archive.writer import ArchiveWriter


class ArchiveBuffer(ArchiveWriter):
    """The archive buffer maintains a list of archive rows in memory."""
    def __init__(self, row_counter):
        """Initialize an empty row buffer and the counter to generate unique
        row identifier.

        Parameters
        ----------
        row_counter: int
            Counter that is used to generate unique internal row identifier.
            The current value of the counter is the value for the next unique
            identifier.
        """
        super(ArchiveBuffer, self).__init__(row_counter)
        self.rows = list()

    def write_archive_row(self, row):
        """Add the given row to the row buffer.

        Parameters
        ----------
        row: histore.archive.row.ArchiveRow
            Row in a new version of a dataset archive.
        """
        self.rows.append(row)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Helper methods for timestamp management."""

from abc import ABCMeta, abstractmethod
from datetime import datetime
from dateutil.parser import isoparse
from dateutil.tz import UTC

import errno
import gzip
import os


# -- Datetime -----------------------------------------------------------------

def to_datetime(timestamp):
    """Converts a timestamp string in ISO format into a datatime object in
    UTC timezone.

    Parameters
    ----------
    timstamp : string
        Timestamp in ISO format

    Returns
    -------
    datetime.datetime
        Datetime object
    """
    try:
        dt = isoparse(timestamp)
        if not dt.tzinfo == UTC:
            dt = dt.astimezone(UTC)
        return dt
    except ValueError:
        raise ValueError('unknown time format {}'.format(timestamp))


def to_localtime(ts):
    """Conbert timestamp in UTC timezone to local time.

    Parameters
    ----------
     ts: datetime.datetime
        Timestamp in UTC format

    Returns
    -------
    datetime.datetime
    """
    return ts.astimezone()


def utc_now():
    """Get the current time in UTC timezone.

    Returns
    -------
    datetime.datetime
    """
    return datetime.now(UTC)


# -- I/O ----------------------------------------------------------------------

def createdir(directory, abs=False):
    """Safely create the given directory path if it does not exist.

    Parameters
    ----------
    directory: string
        Path to directory that is being created.
    abs: boolean, optional
        Return absolute path if true

    Returns
    -------
    string
    """
    # Based on https://stackoverflow.com/questions/273192/
    if not os.path.exists(directory):
        try:
            os.makedirs(directory)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
    if abs:
        return os.path.abspath(directory)
    else:
        return directory


def inputstream(filename, compression=None):
    """Open the given file for input. The compression mide string determines
    which compression algorithm is being used (or no compression if None).

    Parameters
    ----------
    compression: string, default=None
        String representing the compression mode for the output file.

    Returns
    -------
    histore.util.IOStream
    """
    if compression is None:
        return PlainTextFile(open(filename, 'r'))
    elif compression == 'gzip':
        return GZipFile(gzip.open(filename, 'rb'))
    raise ValueError('unknown compression mode {}'.format(compression))


def outputstream(filename, compression=None):
    """Open the given file for output. The compression mide string determines
    which compression algorithm is being used (or no compression if None).

    Parameters
    ----------
    compression: string, default=None
        String representing the compression mode for the output file.

    Returns
    -------
    histore.util.IOStream
    """
    if compression is None:
        return PlainTextFile(open(filename, 'w'))
    elif compression == 'gzip':
        return GZipFile(gzip.open(filename, 'wb'))
    raise ValueError('unknown compression mode {}'.format(compression))


class IOStream(metaclass=ABCMeta):
    def __init__(self, f):
        """Initialize the IO stream.

        Parameters
        ----------
        f: FileObject
            Open file object.
        """
        self.f = f

    def close(self):
        """Close the IO Stream."""
        self.f.close()

    @abstractmethod
    def readline(self):
        """Read sting line from input file. The returned line is stripped of
        any leading or trailing whitespace characters.

        Returns
        ----------
        string
        """
        raise NotImplementedError()

    @abstractmethod
    def writeline(self, line):
        """Write sting line to output file.

        Parameters
        ----------
        line: string
            Output line that is bein written.
        """
        raise NotImplementedError()


class GZipFile(IOStream):
    """IO stream wrapper for gzip compressed files."""
    def __init__(self, f):
        """Initialize the IO stream.

        Parameters
        ----------
        f: FileObject
            Open file object.
        """
        super(GZipFile, self).__init__(f)

    def readline(self):
        """Read sting line from input file. The returned line is stripped of
        any leading or trailing whitespace characters.

        Returns
        ----------
        string
        """
        return self.f.readline().decode('utf8').strip()

    def writeline(self, line):
        """Write sting line to output file.

        Parameters
        ----------
        line: string
            Output line that is bein written.
        """
        self.f.write(str.encode(line))
        self.f.write(b'\n')


class PlainTextFile(IOStream):
    """IO stream for plain text files."""
    def __init__(self, f):
        """Initialize the IO stream.

        Parameters
        ----------
        f: FileObject
            Open file object.
        """
        super(PlainTextFile, self).__init__(f)

    def readline(self):
        """Read sting line from input file. The returned line is stripped of
        any leading or trailing whitespace characters.

        Returns
        ----------
        string
        """
        return self.f.readline().strip()

    def writeline(self, line):
        """Write sting line to output file.

        Parameters
        ----------
        line: string
            Output line that is bein written.
        """
        self.f.write(line)
        self.f.write('\n')



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Nested-merge logic that merges rows in a snapshot document into an existing
dataset archive.
"""


def merge_rows(
    archive, document, version, writer, partial=False, unchanged_cells=None,
    origin=None
):
    """Merge rows in the given archive and database snapshot. Outputs the
    merged rows in the resulting archive to the given archive writer.

    Parameters
    ----------
    archive: histore.archive.base.Archive
        Archive containing previous snapshots of the dataset.
    document: histore.document.base.Document
        Dataset snapshot that is being merged into the archive.
    version: int
        Identifier of the new archive version.
    writer: histore.archive.writer.ArchiveWriter
        Consumer for rows in the new archive version.
    partial: bool, default=False
        Flag indicating whether the given document is a partial document. For
        partial documents missing rows in the archive are considered unchanged
        with respect to the given origin.
    unchanged_cells: set, default=None
        Set of identifier for columns whose cell values are not included in
        the values dictionary for document rows but that remain unchanged with
        respect to the specified source version (origin).
    origin: int, default=None
        Version that the row values originate from. Rows that remain unchanged
        have the timestamp extended for the version of origin.
    """
    # Get reader for the given archive and document.
    arch_reader = archive.reader()
    doc_reader = document.reader()
    # Get the first row for each reader.
    arch_row = arch_reader.next()
    doc_row = doc_reader.next()
    while arch_row is not None and doc_row is not None:
        comp = arch_row.comp(doc_row.key)
        if comp < 0:
            # The row is not present in the document. If the document is
            # a partial document we need to extend the row for the given
            # origin.
            if partial:
                arch_row = arch_row.extend(version=version, origin=origin)
            # Add the row to the new archive version and progress the
            # archive reader.
            writer.write_archive_row(arch_row)
            arch_row = arch_reader.next()
        elif comp > 0:
            # The document row is a new row. Create an archive row from the
            # document row and pass it to the writer. Progress the document
            # reader.
            writer.write_document_row(row=doc_row, version=version)
            doc_row = doc_reader.next()
        else:
            # Merge the archive row and the document row.
            arch_row = arch_row.merge(
                values=doc_row.values,
                pos=doc_row.pos,
                version=version,
                unchanged_cells=unchanged_cells,
                origin=origin
            )
            # Add the merged row to the new archive version and progress
            # both readers.
            writer.write_archive_row(arch_row)
            arch_row = arch_reader.next()
            doc_row = doc_reader.next()
    # Add remaining rows to the archive. Only one of the two conditions can
    # be true; either there are additional archive rows or there are additional
    # document rows.
    # Add remaining archive rows to the new archive version.
    while arch_row is not None:
        # Extend the row for the given origin if the document is partial.
        if partial:
            arch_row = arch_row.extend(version=version, origin=origin)
        writer.write_archive_row(arch_row)
        arch_row = arch_reader.next()
    # Add remaining document rows to the new archive version.
    while doc_row is not None:
        # Outout an archive row created from the document row.
        writer.write_document_row(row=doc_row, version=version)
        doc_row = doc_reader.next()



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Classes representing provenance information for dataset rows in an archive.
Row provenance information is represented as operators that describe the
differences in a single dataset row between two dataset snapshots.
"""

import histore.archive.provenance.base as prov


class DeleteRow(prov.ProvOp):
    """Provenance class representing the row deletion operator."""
    def __init__(self, key):
        """Initialize the operator type in the super class and the row key.

        Parameters
        ----------
        key: scalar or tuple
            Unique row key.
        """
        super(DeleteRow, self).__init__(type=prov.PROV_DELETE, key=key)


class InsertRow(prov.ProvOp):
    """Provenance class representing the row insert operator."""
    def __init__(self, key):
        """Initialize the operator type in the super class and the row key.

        Parameters
        ----------
        key: scalar or tuple
            Unique row key.
        """
        super(InsertRow, self).__init__(type=prov.PROV_INSERT, key=key)


class UpdateRow(prov.ProvOp):
    """Provenance class representing the row update operator. The row update
    operator maintains additional information about potential changes in the
    row position and the row cells.
    """
    def __init__(self, key, cells=None, position=None):
        """Initialize information about the parts of the row that have changed.
        This includes changes to cell values and the row position. Both types
        of changes are optional but at least one of them has to be given.

        Parameters
        ----------
        key: scalar or tuple
            Unique row key.
        cells: dict, default=None
            Dictionaty mapping column identifier to changed cell values.
        position: histore.archive.provenance.value.UpdateValue, default=None
            Updated value representing the old and new position of the row in
            the dataset.

        Raises
        ------
        ValueError
        """
        # Raise an error if both arguments are None
        if not cells and not position:
            raise ValueError('no row changes given')
        super(UpdateRow, self).__init__(type=prov.PROV_UPDATE, key=key)
        self.cells = cells
        self.position = position

    def updated_cells(self):
        """Get dictionary mapping column identifier to updated cell values.

        Returns
        -------
        dict
        """
        return self.cells

    def updated_position(self):
        """Get updated row position.

        Returns
        -------
        histore.archive.provenance.value.UpdateValue
        """
        return self.position



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Writer for archives that are materialized as Json files on the file
system.
"""

import json
import numpy as np

from histore.archive.serialize.default import DefaultSerializer
from histore.archive.writer import ArchiveWriter

import histore.util as util


class ArchiveFileWriter(ArchiveWriter):
    """Archive writer that outputs rows in an archive as Json serialized rows
    in a text file. Each row is stired ina separate line in the text file. The
    output file is a Json array. The first and the last row of the file open
    and close the array.
    """
    def __init__(
        self, filename, row_counter=0, serializer=None, compression=None
    ):
        """Initialize the output file, row counter, and the serializer that is
        being used.

        Parameters
        ----------
        filename: string
            Path to the output file.
        row_counter: int, default=0
            Counter that is used to generate unique internal row identifier.
            The current value of the counter is the value for the next unique
            identifier.
        serializer: histore.archive.serialize.ArchiveSerializer, default=None
            Implementation of the archive serializer interface that is used to
            serialize rows that are written to file.
        compression: string, default=None
            String representing the compression mode for the output file.
        """
        super(ArchiveFileWriter, self).__init__(row_counter)
        # Use the default serializer if no serializer was given
        self.serializer = serializer if serializer else DefaultSerializer()
        # Open output file for writing.
        self.fout = util.outputstream(filename, compression=compression)
        # Write buffer to keep track of the last row that is being written.
        self.buffer = None

    def close(self):
        """Write the last row to the output file and close the output array and
        the output file.
        """
        # Write last row in to output buffer.
        self.write_buffer()
        # Close Json array and the output file.
        self.fout.writeline(']')
        self.fout.close()

    def write_archive_row(self, row):
        """Add the given row to the output file.

        Parameters
        ----------
        row: histore.archive.row.ArchiveRow
            Row in a new version of a dataset archive.
        """
        self.write_buffer(row)

    def write_buffer(self, row=None):
        """Write the archive row in the internal buffer to the output file.
        Replace the buffer with the given (next output row).

        Parameters
        ----------
        row: histore.archive.row.ArchiveRow, default=None
            Next row in the output stream. This row will be kept in the
            internal buffer and the previous row is being written to the
            output file.
        """
        # Depending on whether there is a previous row in the buffer we either
        # output that row or open the output array.
        if self.buffer is not None:
            text = self.serializer.serialize_row(self.buffer)
            line = json.dumps(text, cls=NumpyEncoder)
            if row is not None:
                line += ','
        else:
            line = '['
        self.fout.writeline(line)
        # Buffer the given row
        self.buffer = row


# -- Helper classes -----------------------------------------------------------

class NumpyEncoder(json.JSONEncoder):
    """Json encoder that handles numpy data types from pandas data frames.

    Based on https://stackoverflow.com/questions/50916422.
    """
    def default(self, obj):
        """Convert numpy data types to default Python types."""
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyEncoder, self).default(obj)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Required packages for install, test, docs, and tests."""

import os
import re

from setuptools import setup, find_packages


install_requires = [
    'pandas>=1.0.0',
    'pyyaml'
]


tests_require = [
    'coverage>=4.0',
    'pytest',
    'pytest-cov',
    'tox'
]


extras_require = {
    'docs': [
        'Sphinx',
        'sphinx-rtd-theme'
    ],
    'tests': tests_require,
}


# Get the version string from the version.py file in the histore package.
# Based on:
# https://stackoverflow.com/questions/458550
with open(os.path.join('histore', 'version.py'), 'rt') as f:
    filecontent = f.read()
match = re.search(r"^__version__\s*=\s*['\"]([^'\"]*)['\"]", filecontent, re.M)
if match is not None:
    version = match.group(1)
else:
    raise RuntimeError('unable to find version string in %s.' % (filecontent,))


# Get long project description text from the README.rst file
with open('README.rst', 'rt') as f:
    readme = f.read()


setup(
    name='histore',
    version=version,
    description='Library for maintaining evolving tabular data sets',
    long_description=readme,
    long_description_content_type='text/x-rst',
    keywords='data versioning',
    url='https://github.com/heikomuller/histore',
    author='Heiko Mueller',
    author_email='heiko.muller@gmail.com',
    license='New BSD',
    license_file='LICENSE',
    packages=find_packages(exclude=('tests',)),
    include_package_data=True,
    extras_require=extras_require,
    tests_require=tests_require,
    install_requires=install_requires,
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python'
    ]
)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Rows in a dataset snapshot document. Each row has an identifier, the index
position of the row in the dataset, and a list of cell values for all dataset
columns.
"""


class DocumentRow(object):
    """Each row in snapshot document has a row identifier, an index position
    and the set of cell values for columns in the dataset snapshot. Row
    identifier may either be derived from a primary key column(s) or be an
    integer from the data frame row index. Row identifier may be negative (and
    not unique) indicating new rows for which unique identifier need to be
    generated.
    """
    def __init__(self, key, pos, values):
        """Initialize the row identifier, the index positions, and the cell
        values. Cell values are provided as a dictionary that maps column
        identifier (from the dataset archive schema) to cell values.

        Parameters
        ----------
        key: int, string, or tuple
            Derived row identifier.
        pos: int
            Index position for the row in the dataset.
        values: dict(int, scalar)
            Mapping of column identifier from the archive schema to cell values
            in the corresponding data frame row cells.
        """
        self.key = key
        self.pos = pos
        self.values = values

    def __repr__(self):
        """Unambiguous string representation of the document row.

        Returns
        -------
        string
        """
        return '<DocumentRow(key={}, pos={}, values={})'.format(
            self.key,
            self.pos,
            self.values
        )



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Descriptor classes that can be used to print summary information for a set
of provenance operators.
"""

from abc import ABCMeta, abstractmethod


class ProvDescriptor(metaclass=ABCMeta):
    """Abstract provenance summary printer."""
    @abstractmethod
    def describe(self, provenance):
        """Print summary of the maintained provenance information."""
        raise NotImplementedError()


class DefaultDescribe(ProvDescriptor):
    """Default provenance summary printer. Outputs counts for inserted,
    deleted, and updated objects.
    """
    def describe(self, provenance):
        """Print summary of the maintained provenance information."""
        print('Inserted: {}'.format(len(provenance.insert())))
        print('Deleted : {}'.format(len(provenance.delete())))
        print('Updated : {}'.format(len(provenance.update())))


class RowProvDescribe(ProvDescriptor):
    """Print provenance summary information about changed rows."""
    def describe(self, provenance):
        """Print summary of the maintained provenance information."""
        updated_rows, updated_values, moved_rows = 0, 0, 0
        for upd_row in provenance.update():
            if upd_row.updated_position():
                moved_rows += 1
            if upd_row.updated_cells():
                updated_rows += 1
                updated_values += len(upd_row.updated_cells())
        print('Inserted Rows    : {}'.format(len(provenance.insert())))
        print('Deleted Rows     : {}'.format(len(provenance.delete())))
        print('Moved Rows       : {}'.format(moved_rows))
        print('Updated Rows     : {}'.format(updated_rows))
        print('Updated Values   : {}'.format(updated_values))


class SchemaProvDescribe(ProvDescriptor):
    """Print provenance summary information about schema changes."""
    def describe(self, provenance):
        """Print summary of the maintained provenance information."""
        moved_cols, renamed_cols = 0, 0
        for upd_row in provenance.update():
            if upd_row.updated_position():
                moved_cols += 1
            if upd_row.updated_name():
                renamed_cols += 1
        print('Inserted Columns : {}'.format(len(provenance.insert())))
        print('Deleted Columns  : {}'.format(len(provenance.delete())))
        print('Moved Columns    : {}'.format(moved_cols))
        print('Renamed Columns  : {}'.format(renamed_cols))



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""The archive store intruduces a layer of abstraction between the archive
object and the way in which archives are managed and maintained.
"""

from abc import ABCMeta, abstractmethod


class ArchiveStore(metaclass=ABCMeta):
    """Abstract class that defines the interface to read and write dataset
    archive information.
    """
    @abstractmethod
    def commit(self, schema, writer, snapshots):
        """Commit a new version of the dataset archive. The modified components
        of the archive are given as the three arguments of this method.

        Parameters
        ----------
        schema: histore.archive.schema.ArchiveSchema
            Schema history for the new archive version.
        writer: histore.archive.writer.ArchiveWriter
            Instance of the archive writer class returned by this store that
            was used to output the rows of the new archive version.
        snapshots: histore.archive.snapshot.SnapshotListing
            Modified list of snapshots in the new archive. The new archive
            version is the last entry in the list.
        """
        raise NotImplementedError()

    @abstractmethod
    def is_empty(self):
        """True if the archive does not contain any snapshots yet.

        Returns
        -------
        bool
        """
        raise NotImplementedError()

    @abstractmethod
    def get_reader(self):
        """Get the row reader for this archive.

        Returns
        -------
        histore.archive.reader.ArchiveReader
        """
        raise NotImplementedError()

    @abstractmethod
    def get_schema(self):
        """Get the schema history for the archived dataset.

        Returns
        -------
        histore.archive.schema.ArchiveSchema
        """
        raise NotImplementedError()

    @abstractmethod
    def get_snapshots(self):
        """Get listing of all snapshots in the archive.

        Returns
        -------
        histore.archive.snapshot.SnapshotListing
        """
        raise NotImplementedError()

    @abstractmethod
    def get_writer(self):
        """Get a writer for a new version of the archive.

        Returns
        -------
        histore.archive.writer.ArchiveWriter
        """
        raise NotImplementedError()



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Archives are collections of snapshots of an evolving dataset."""

import pandas as pd

from histore.archive.provenance.archive import SnapshotDiff
from histore.archive.reader import RowIndexReader
from histore.archive.schema import MATCH_ID, MATCH_IDNAME
from histore.archive.store.fs.base import ArchiveFileStore
from histore.archive.store.mem.base import VolatileArchiveStore
from histore.document.base import PartialDocument, PKDocument, RIDocument

import histore.archive.merge as nested_merge


class Archive(object):
    """An archive maintains a list of snapshot handles. All snapshots are
    expected to follow the same document schema.

    Archives are represented as trees. The root of the tree is maintained by
    an archive store. The store provides a layer of abstraction such that the
    archive object does not have to deal with the different ways in which
    archives are managed by different systems.
    """
    def __init__(self, store=None, primary_key=None):
        """Initialize the associated archive store and the optional primary
        key columns that are used to generate row identifier. If no primary
        key is specified the row index for committed data frame is used to
        generate identifier for archive rows.

        Parameters
        ----------
        store: histore.archive.store.base.ArchiveStore, default=None
            Associated archive store that is used to maintain all archive
            information. Uses the volatile in-memory store by default.
        primary_key: string or list
            Column(s) that are used to generate identifier for snapshot rows.
        """
        self.primary_key = primary_key
        self.store = store if store is not None else VolatileArchiveStore()

    def checkout(self, version):
        """Access a dataset snapshot in the archive. Retrieves the datset that
        was commited with the given version identifier. Raises an error if the
        version identifier is unknown.

        Parameters
        ----------
        version: int
            Unique version identifier.

        Returns
        -------
        pandas.DataFrame

        Raises
        ------
        ValueError
        """
        # Ensure that the version exists in the snapshot index.
        if not self.snapshots().has_version(version):
            raise ValueError('unknown version {}'.format(version))
        # Get dataset schema at the given version.
        columns = self.schema().at_version(version)
        colids = [c.colid for c in columns]
        # Get the row values and their position.
        rows = list()
        reader = self.reader()
        while reader.has_next():
            row = reader.next()
            if row.timestamp.contains(version):
                pos, vals = row.at_version(version, colids, raise_error=False)
                rows.append((row.rowid, pos, vals))
        # Sort rows in ascending order.
        rows.sort(key=lambda r: r[1])
        # Create data frame for the retrieved snapshot.
        data, rowindex = list(), list()
        for rowid, _, vals in rows:
            data.append(vals)
            rowindex.append(rowid)
        return pd.DataFrame(data=data, index=rowindex, columns=columns)

    def commit(
        self, df, description=None, valid_time=None, matching=MATCH_IDNAME,
        renamed=None, renamed_to=True, partial=False, origin=None
    ):
        """Commit a new snapshot to the dataset archive. The given data frame
        represents the dataset snapshot that is being merged into the archive.
        The data frame may represent a complete snapshot of the data or only
        a partial snapshot. In the latter case, all columns and rows from the
        snapshot that the data frame originated from (origin) are considered
        unchanged.

        Matching of snapshot schema columns to archive columns can either be
        done by name or by their unique identifier (if present). When matching
        by name a snapshot version of origin is used to match against. Data
        frames or snapshot versions with duplicate column names cannot be used
        for matching columns by name.

        If no origin is specified, the last commited snapshot is assumed as the
        default snapshot of origin. If this is the first snapshot in the
        archive the partial flag cannot be True (will raise a ValueError).

        Returns the descriptor of the merged snapshot in the new version of
        the archive.

        Parameters
        ----------
        df: pandas.DataFrame
            Data frame representing the dataset snapshot that is being merged
            into the archive.
        description: string, default=None
            Optional user-provided description for the snapshot.
        valid_time: datetime.datetime
            Timestamp when the snapshot was first valid. A snapshot is valid
            until the valid time of the next snapshot in the archive.
        matching: string, default='idname'
            Match mode for columns. Excepts one of three modes:
            - idonly: The columns in the schema of the comitted data frame are
            matched against columns in the archive schema by their identifier.
            Assumes that columns in the data frame schema are instances of the
            class histore.document.schema.Column.
            - nameonly: Columns in the commited data frame schema are matched
            by name against the columns in the schema of the snapshot that is
            identified by origin.
            - idname: Match columns of type histore.document.schema.Column
            first against the columns in the archive schema. Match remaining
            columns by name against the schema of the snapshot that is
            identified by origin.
        renamed: dict, default=None
            Optional mapping of columns that have been renamed. Maps the new
            column name to the original name.
        renamed_to: bool, default=True
            Flag that determines the semantics of the mapping in the renamed
            dictionary. By default a mapping from the original column name
            (i.e., the dictionary key) to the new column name (the dictionary
            value) is assumed. If the flag is False a mapping from the new
            column name to the original column name is assumed.
        partial: bool, default=False
            If True the given snapshot is assumed to be partial. All columns
            from the snapshot schema that is specified by origin that are not
            matched by any column in the snapshot schema are assumed to be
            unchanged. All rows from the orignal snapshot that are not in the
            given snapshot are also assumed to be unchnged.
        origin: int, default=None
            Version identifier of the original column against which the given
            column list is matched.

        Returns
        -------
        histore.archive.snapshot.Snapshot

        Raises
        ------
        ValueError
        """
        # Ensure that partial is not set for an empty archive.
        if partial and self.is_empty():
            raise ValueError('merge partial snapshot into empty archive')
        # Use the last commited version as origin if matching columns by name
        # or if a partial data frame is commited and origin is None.
        if (matching != MATCH_ID or partial) and origin is None:
            last_snapshot = self.snapshots().last_snapshot()
            if last_snapshot:
                origin = last_snapshot.version
        # Get a modified snapshot list where the last entry represents the
        # new snapshot.
        snapshots = self.snapshots().append(
            valid_time=valid_time,
            description=description
        )
        version = snapshots.last_snapshot().version
        # Merge the new snapshot schema with the current archive schema.
        schema, matched_columns, unchanged_columns = self.schema().merge(
            columns=list(df.columns),
            version=version,
            matching=matching,
            renamed=renamed,
            renamed_to=renamed_to,
            partial=partial,
            origin=origin
        )
        # Create snapshot document for the given data frame. The document type
        # depends on how row identifier are generated.
        if self.primary_key is not None:
            doc = PKDocument(
                df=df,
                schema=matched_columns,
                primary_key=self.primary_key
            )
        else:
            doc = RIDocument(df=df, schema=matched_columns)
        # If the data frame is partial we need to adjust the positions of the
        # rows in the document.
        if partial:
            doc = PartialDocument(
                doc=doc,
                row_index=RowIndexReader(reader=self.reader(), version=origin)
            )
        # Merge document rows into the archive.
        writer = self.store.get_writer()
        nested_merge.merge_rows(
            archive=self,
            document=doc,
            version=version,
            writer=writer,
            partial=partial,
            unchanged_cells=[c.colid for c in unchanged_columns],
            origin=origin
        )
        # Commit all changes to the associated archive store.
        self.store.commit(schema=schema, writer=writer, snapshots=snapshots)
        # Return descriptor for the created snapshot.
        return snapshots.last_snapshot()

    def diff(self, original_version, new_version):
        """Get provenance information representing the difference between two
        dataset snapshots.

        Parameters
        ----------
        original_version: int
            Unique identifier for the original version.
        new_version: int
            Unique identifier for the version that the original version is
            compared against.

        Returns
        -------
        histore.archive.provenance.archive.SnapshotDiff
        """
        prov = SnapshotDiff()
        # Get changes in the dataset schema.
        schema_diff = prov.schema()
        for colid, column in self.schema().columns.items():
            col_prov = column.diff(original_version, new_version)
            if col_prov is not None:
                schema_diff.add(col_prov)
        # Get changes in dataset rows.
        rows_diff = prov.rows()
        reader = self.reader()
        row_count = 0
        while reader.has_next():
            row = reader.next()
            row_prov = row.diff(original_version, new_version)
            if row_prov is not None:
                rows_diff.add(row_prov)
            row_count += 1
        return prov

    def is_empty(self):
        """True if the archive does not contain any snapshots yet.

        Returns
        -------
        bool
        """
        return self.store.is_empty()

    def reader(self):
        """Get the row reader for this archive.

        Returns
        -------
        histore.archive.reader.ArchiveReader
        """
        return self.store.get_reader()

    def schema(self):
        """Get the schema history for the archived dataset.

        Returns
        -------
        histore.archive.schema.ArchiveSchema
        """
        return self.store.get_schema()

    def snapshots(self):
        """Get listing of all snapshots in the archive.

        Returns
        -------
        histore.archive.snapshot.SnapshotListing
        """
        return self.store.get_snapshots()


class PersistentArchive(Archive):
    """Archive that persists all dataset snapshots in files on the file system.
    All the data is maintained within a given directory on the file system.

    This class is a wrapper for an archive with a persistent archive store. It
    uses the default file system store.
    """
    def __init__(
        self, basedir, replace=False, serializer=None, compression=None,
        primary_key=None
    ):
        """Initialize the associated archive store and the optional primary
        key columns that are used to generate row identifier. If no primary
        key is specified the row index for committed data frame is used to
        generate identifier for archive rows.

        Parameters
        ----------
        basedir: string
            Path to the base directory for archive files. If the directory does
            not exist it will be created.
        replace: boolean, default=False
            Do not read existing files in the base directory to initialize the
            archive. Treat the archive as being empty instead if True.
        serializer: histore.archive.serialize.ArchiveSerializer, default=None
            Implementation of the archive serializer interface that is used to
            serialize rows that are written to file.
        compression: string, default=None
            String representing the compression mode. Only te data file will be
            compressed. the metadata file is always storesd as plain text.
        primary_key: string or list
            Column(s) that are used to generate identifier for snapshot rows.
        """
        super(PersistentArchive, self).__init__(
            store=ArchiveFileStore(
                basedir=basedir,
                replace=replace,
                serializer=serializer,
                compression=compression
            ),
            primary_key=primary_key
        )



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

from abc import ABCMeta, abstractmethod


class ArchiveReader(metaclass=ABCMeta):
    """Reader for rows in a dataset archive. Reads rows in ascending order of
    their identifier.
    """
    @abstractmethod
    def has_next(self):
        """Test if the reader has more rows to read. If True the next() method
        will return the next row. Otherwise, the next() method will return
        None.

        Returns
        -------
        bool
        """
        raise NotImplementedError()

    @abstractmethod
    def next(self):
        """Read the next row in the dataset archive. Returns None if the end of
        the archive rows has been reached.

        Returns
        -------
        histore.archive.row.ArchiveRow
        """
        raise NotImplementedError()


class RowIndexReader(object):
    """Reader for row index information for a given snapshot version in a
    dataset archive. The row information is read in ascending order of the
    row identifier in the archive.
    """
    def __init__(self, reader, version):
        """Initialize the archive reader and the version for which the index is
        being read.

        Parameters
        ----------
        reader: histore.archive.reader.ArchiveReader
            Reader for a dataset archive.
        version: int
            Identifier of version for which the row index is being read.
        """
        self.reader = reader
        self.version = version

    def next(self):
        """Get information for the next row in the archive. The result is a
        tuple of row identifier and position. If the end of the archive has
        been reached the result is None.

        Returns
        -------
        tuple
        """
        while self.reader.has_next():
            row = self.reader.next()
            if row.timestamp.contains(self.version):
                return (row.key, row.pos.at_version(self.version))
